"""Module shellInfoDialog

Implements shell configuration dialog.

"""

import os
import sys
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa

import pyzo
from pyzo.core.pyzoLogging import print
from pyzo.core.kernelbroker import KernelInfo
from pyzo import translate


## Implement widgets that have a common interface


class ShellInfoLineEdit(QtWidgets.QLineEdit):
    def setTheText(self, value):
        self.setText(value)

    def getTheText(self):
        return self.text()


class ShellInfo_name(ShellInfoLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.editingFinished.connect(self.onValueChanged)
        t = translate("shell", "name ::: The name of this configuration.")
        self.setPlaceholderText(t.tt)

    def setTheText(self, value):
        super().setTheText(value)
        self.onValueChanged()

    def onValueChanged(self):
        self.parent().parent().parent().setShellconfigTitle(self.getTheText())


class ShellInfo_exe(QtWidgets.QComboBox):
    def __init__(self, *args):
        super().__init__(*args)

    def _interpreterName(self, p):
        if p.is_conda:
            return "{}  [v{}, conda]".format(p.path, p.version)
        else:
            return "{}  [v{}]".format(p.path, p.version)

    def setTheText(self, value):
        # Init
        self.clear()
        self.setEditable(True)
        self.setInsertPolicy(self.InsertPolicy.InsertAtTop)

        # Get known interpreters from shellDialog (which are sorted by version)
        shellDialog = self
        while not isinstance(shellDialog, ShellInfoDialog):
            shellDialog = shellDialog.parent()
        interpreters = shellDialog.interpreters
        exes = [p.path for p in interpreters]

        # Hande current value
        if value in exes:
            value = self._interpreterName(interpreters[exes.index(value)])
        else:
            self.addItem(value)

        # Add all found interpreters
        for p in interpreters:
            self.addItem(self._interpreterName(p))

        # Set current text
        self.setEditText(value)

    def getTheText(self):
        # return self.currentText().split('(')[0].rstrip()
        value = self.currentText()
        if value.endswith("]") and "[" in value:
            value = value.rsplit("[", 1)[0]
        return value.strip()


class ShellInfo_ipython(QtWidgets.QCheckBox):
    def __init__(self, parent):
        super().__init__(parent)
        t = translate("shell", "ipython ::: Use IPython shell if available.")
        self.setText(t.tt)
        self.setChecked(False)

    def setTheText(self, value):
        self.setChecked(value.lower() not in ("", "no", "false"))

    def getTheText(self):
        return "yes" if self.isChecked() else "no"


class ShellInfo_gui(QtWidgets.QComboBox):
    # GUI names
    GUIS = [
        ("None", "no GUI support"),
        ("Auto", "Use what is available (recommended)"),
        ("Asyncio", "Python's builtin event loop"),
        ("PySide6", "LGPL licensed wrapper to Qt6"),
        ("PySide2", "LGPL licensed wrapper to Qt5"),
        ("PyQt6", "GPL/commercial licensed wrapper to Qt6"),
        ("PyQt5", "GPL/commercial licensed wrapper to Qt5"),
        ("Tornado", "Tornado asynchronous networking library"),
        ("Tk", "Tk widget toolkit"),
        ("WX", "wxPython"),
        ("FLTK", "The fast light toolkit"),
        ("GTK", "GIMP Toolkit"),
    ]

    # GUI descriptions

    def setTheText(self, value):
        # Process value
        value = value.upper()

        # Set options
        ii = 0
        self.clear()
        for i, (gui, des) in enumerate(self.GUIS):
            if value == gui.upper():
                ii = i
            self.addItem("{}  -  {}".format(gui, des))

        # Set current text
        self.setCurrentIndex(ii)

    def getTheText(self):
        text = self.currentText().lower()
        return text.partition("-")[0].strip()


class ShellinfoWithSystemDefault(QtWidgets.QVBoxLayout):
    DISABLE_SYSTEM_DEFAULT = sys.platform == "darwin"
    SYSTEM_VALUE = ""

    def __init__(self, parent, widget):
        # Do not pass parent, because is a sublayout
        super().__init__()

        # Layout
        self.setSpacing(1)
        self.addWidget(widget)

        # Create checkbox widget
        if not self.DISABLE_SYSTEM_DEFAULT:
            t = translate("shell", "Use system default")
            self._check = QtWidgets.QCheckBox(t, parent)
            self._check.stateChanged.connect(self.onCheckChanged)
            self.addWidget(self._check)

        # The actual value of this shell config attribute
        self._value = ""

        # A buffered version, so that clicking the text box does not
        # remove the value at once
        self._bufferedValue = ""

    def onEditChanged(self):
        if self.DISABLE_SYSTEM_DEFAULT or not self._check.isChecked():
            self._value = self.getWidgetText()

    def onCheckChanged(self, state):
        if state:
            self._bufferedValue = self._value
            self.setTheText(self.SYSTEM_VALUE)
        else:
            self.setTheText(self._bufferedValue)

    def setTheText(self, value):
        if self.DISABLE_SYSTEM_DEFAULT:
            # Just set the value
            self._edit.setReadOnly(False)
            self.setWidgetText(value)

        elif value != self.SYSTEM_VALUE:
            # Value given, enable edit
            self._check.setChecked(False)
            self._edit.setReadOnly(False)
            # Set the text
            self.setWidgetText(value)

        else:
            # Use system default, disable edit widget
            self._check.setChecked(True)
            self._edit.setReadOnly(True)
            # Set text using system environment
            self.setWidgetText(None)

        # Store value
        self._value = value

    def getTheText(self):
        return self._value


class ShellInfo_pythonPath(ShellinfoWithSystemDefault):
    SYSTEM_VALUE = "$PYTHONPATH"

    def __init__(self, parent):
        # Create sub-widget
        self._edit = QtWidgets.QPlainTextEdit(parent)
        self._edit.zoomOut(1)
        self._edit.setMaximumHeight(60)
        self._edit.setMinimumWidth(200)
        self._edit.textChanged.connect(self.onEditChanged)

        # Instantiate
        super().__init__(parent, self._edit)

    def getWidgetText(self):
        return self._edit.toPlainText()

    def setWidgetText(self, value=None):
        if value is None:
            pp = os.environ.get("PYTHONPATH", "")
            pp = pp.replace(os.pathsep, "\n").strip()
            value = "$PYTHONPATH:\n{}\n".format(pp)
        self._edit.setPlainText(value)


class ShellInfo_startupScript(QtWidgets.QVBoxLayout):
    DISABLE_SYSTEM_DEFAULT = sys.platform == "darwin"
    SYSTEM_VALUE = "$PYTHONSTARTUP"
    RUN_BEFORE_AFTER_GUI_TEXT = (
        "... Python code to run before integrating the GUI ...\n"
        "# AFTER_GUI\n"
        "... Python code to run after integrating the GUI ..."
    )

    def __init__(self, parent):
        # Do not pass parent, because is a sublayout
        super().__init__()

        # Create sub-widget
        self._edit1 = QtWidgets.QLineEdit(parent)
        self._edit1.textEdited.connect(self.onEditChanged)
        if sys.platform.startswith("win"):
            self._edit1.setPlaceholderText(r"C:\path\to\script.py")
        else:
            self._edit1.setPlaceholderText("/path/to/script.py")
        #
        self._edit2 = QtWidgets.QPlainTextEdit(parent)
        self._edit2.zoomOut(1)
        self._edit2.setMinimumWidth(200)
        self._edit2.textChanged.connect(self.onEditChanged)
        self._edit2.setPlaceholderText(self.RUN_BEFORE_AFTER_GUI_TEXT)

        # Layout
        self.setSpacing(1)
        self.addWidget(self._edit1)
        self.addWidget(self._edit2)

        # Create radio widget for system default
        t = translate("shell", "Use system default")
        self._radio_system = QtWidgets.QRadioButton(t, parent)
        self._radio_system.toggled.connect(self.onCheckChanged)
        self.addWidget(self._radio_system)
        if self.DISABLE_SYSTEM_DEFAULT:
            self._radio_system.hide()

        # Create radio widget for file
        t = translate("shell", "File to run at startup")
        self._radio_file = QtWidgets.QRadioButton(t, parent)
        self._radio_file.toggled.connect(self.onCheckChanged)
        self.addWidget(self._radio_file)

        # Create radio widget for code
        t = translate("shell", "Code to run at startup")
        self._radio_code = QtWidgets.QRadioButton(t, parent)
        self._radio_code.toggled.connect(self.onCheckChanged)
        self.addWidget(self._radio_code)

        # The actual value of this shell config attribute
        self._value = ""

        # A buffered version, so that clicking the text box does not
        # remove the value at once
        self._valueFile = ""
        self._valueCode = "\n"

    def onEditChanged(self):
        if self._radio_file.isChecked():
            self._value = self._valueFile = self._edit1.text().strip()
        elif self._radio_code.isChecked():
            # ensure newline!
            self._value = self._valueCode = self._edit2.toPlainText().strip() + "\n"

    def onCheckChanged(self, state):
        if self._radio_system.isChecked():
            self.setWidgetText(self.SYSTEM_VALUE)
        elif self._radio_file.isChecked():
            self.setWidgetText(self._valueFile)
        elif self._radio_code.isChecked():
            self.setWidgetText(self._valueCode)

    def setTheText(self, value):
        self.setWidgetText(value, True)
        self._value = value

    def setWidgetText(self, value, init=False):
        self._value = value

        if value == self.SYSTEM_VALUE and not self.DISABLE_SYSTEM_DEFAULT:
            # System default
            if init:
                self._radio_system.setChecked(True)
            pp = os.environ.get("PYTHONSTARTUP", "").strip()
            if pp:
                value = '$PYTHONSTARTUP: "{}"'.format(pp)
            else:
                value = "$PYTHONSTARTUP: None"
            #
            self._edit1.setReadOnly(True)
            self._edit1.show()
            self._edit2.hide()
            self._edit1.setText(value)

        elif "\n" not in value:
            # File
            if init:
                self._radio_file.setChecked(True)
            self._edit1.setReadOnly(False)
            self._edit1.show()
            self._edit2.hide()
            self._edit1.setText(value)

        else:
            # Code
            if init:
                self._radio_code.setChecked(True)
            self._edit1.hide()
            self._edit2.show()
            if not value.strip():
                value = ""  # this will display the placeholder text
            self._edit2.setPlainText(value)

    def getTheText(self):
        return self._value


class ShellInfo_startDir(ShellInfoLineEdit):
    def __init__(self, parent):
        super().__init__(parent)
        if sys.platform.startswith("win"):
            self.setPlaceholderText(r"C:\path\to\your\python\modules")
        else:
            self.setPlaceholderText("/path/to/your/python/modules")


class ShellInfo_argv(ShellInfoLineEdit):
    def __init__(self, parent):
        super().__init__(parent)
        self.setPlaceholderText('arg1 arg2 "arg with spaces"')


class ShellInfo_environ(QtWidgets.QPlainTextEdit):
    EXAMPLE = (
        "PYZO_PROCESS_EVENTS_WHILE_DEBUGGING=1\n"
        "PYZO_INTROSPECT_IN_MAIN_THREAD=1\n"
        "EXAMPLE_VAR1=value1"
    )

    def __init__(self, parent):
        super().__init__(parent)
        self.zoomOut(1)
        self.setPlaceholderText(self.EXAMPLE)

    def _cleanText(self, txt):
        return "\n".join([line.strip() for line in txt.splitlines()])

    def setTheText(self, value):
        value = self._cleanText(value)
        self.setPlainText(value)

    def getTheText(self):
        value = self.toPlainText()
        value = self._cleanText(value)
        return value


## The dialog class and container with pages


class ShellInfoPage(QtWidgets.QScrollArea):
    INFO_KEYS = [
        translate("shell", "name ::: The name of this configuration."),
        translate("shell", "exe ::: The Python executable."),
        translate("shell", "ipython ::: Use IPython shell if available."),
        translate(
            "shell",
            "gui ::: The GUI toolkit to integrate (for interactive plotting, etc.).",
        ),
        translate(
            "shell",
            "pythonPath ::: A list of directories to search for modules and packages. Write each path on a new line, or separate with the default seperator for this OS.",
        ),
        translate(
            "shell",
            "startupScript ::: The script to run at startup (not in script mode).",
        ),
        translate("shell", "startDir ::: The start directory (not in script mode)."),
        translate("shell", "argv ::: The command line arguments (sys.argv)."),
        translate("shell", "environ ::: Extra environment variables (os.environ)."),
    ]

    def __init__(self, parent):
        super().__init__(parent)

        # Init the scroll area
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setWidgetResizable(True)
        self.setFrameShape(QtWidgets.QFrame.Shape.NoFrame)

        # Create widget and a layout
        self._content = QtWidgets.QWidget(parent)
        self._formLayout = QtWidgets.QFormLayout(self._content)

        # Collect classes of widgets to instantiate
        classes = []
        for t in self.INFO_KEYS:
            className = "ShellInfo_" + t.key
            cls = globals()[className]
            classes.append((t, cls))

        # Instantiate all classes
        self._shellInfoWidgets = {}
        for t, cls in classes:
            # Instantiate and store
            instance = cls(self._content)
            self._shellInfoWidgets[t.key] = instance
            # Create label
            label = QtWidgets.QLabel(t, self._content)
            label.setToolTip(t.tt)
            # Add to layout
            self._formLayout.addRow(label, instance)

        # Apply layout
        self._formLayout.setSpacing(15)
        self._content.setLayout(self._formLayout)
        self.setWidget(self._content)

        # Create a list item for switching pages via a QListWidget
        self._listItem = QtWidgets.QListWidgetItem()

    @property
    def listItem(self):
        return self._listItem

    def setShellconfigTitle(self, name):
        self._listItem.setText(name)

    def setInfo(self, info=None):
        """Set the shell info struct, and use it to update the widgets.
        Not via init, because this function also sets the page name.
        """

        # If info not given, use default as specified by the KernelInfo struct
        if info is None:
            info = KernelInfo()
            # Name
            n = self.parent().count()
            if n > 1:
                info.name = "Shell config {}".format(n)

        # Store info
        self._info = info

        # Set widget values according to info
        try:
            for key in info:
                widget = self._shellInfoWidgets.get(key, None)
                if widget is not None:
                    widget.setTheText(info[key])

        except Exception as why:
            print("Error setting info in shell config:", why)
            print(info)

    def getInfo(self):
        info = self._info

        # Set struct values according to widgets
        try:
            for key, widget in self._shellInfoWidgets.items():
                info[key] = widget.getTheText()

        except Exception as why:
            print("Error getting info in shell config:", why)
            print(info)

        # Return the original (but modified) ssdf Dict object
        return info


class MyQListWidgetDragDropCopy(QtWidgets.QListWidget):
    """
    This derived class is only necessary to make drag'n'drop
    copying of list entries with signalling of original and new
    items possible.
    """

    dragDropRowCopied = QtCore.Signal(int, int)

    def __init__(self, *args):
        super().__init__(*args)

        # Setup drag'n'drop for the list
        self.setDragEnabled(True)
        self.setDefaultDropAction(QtCore.Qt.DropAction.MoveAction)
        self.setDragDropMode(self.DragDropMode.DragDrop)
        self._indexDragStart = None
        self._itemsAtDragStart = None

    def dragEnterEvent(self, event):
        if event.source() is self:
            self._indexDragStart = self.row(self.currentItem())
            self._itemsAtDragStart = [self.item(i) for i in range(self.count())]
            event.acceptProposedAction()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.source() is self:
            super().dropEvent(event)
        else:
            event.ignore()
        self._indexDragStart = None
        self._itemsAtDragStart = None

    def rowsInserted(self, parentIndex, start, end):
        """find out which item was copied and where the new item was inserted"""
        super().rowsInserted(parentIndex, start, end)
        indsOfNewItems = [
            i for i in range(self.count()) if self.item(i) not in self._itemsAtDragStart
        ]
        if len(indsOfNewItems) == 1:
            (indNewRow,) = indsOfNewItems
            indOriginalRow = self._indexDragStart
            if indNewRow <= indOriginalRow:
                indOriginalRow += 1
            self.dragDropRowCopied.emit(indOriginalRow, indNewRow)


class ShellInfoDialog(QtWidgets.QDialog):
    """Dialog to edit the shell configurations."""

    def __init__(self, *args):
        super().__init__(*args)
        self.setModal(True)

        # Set title
        self.setWindowTitle(pyzo.translate("shell", "Shell configurations"))

        # Create a page-stack for all configs and a list for switching pages
        self._stack = QtWidgets.QStackedWidget()
        self._list = MyQListWidgetDragDropCopy()
        self._list.currentItemChanged.connect(self._onListItemChanged)
        self._list.dragDropRowCopied.connect(self._onDragDropRowCopied)

        # Get known interpreters (sorted them by version)
        # Do this here so we only need to do it once ...
        from pyzo.util.interpreters import get_interpreters

        self.interpreters = list(reversed(get_interpreters("2.7")))

        cfgSelLayout = QtWidgets.QVBoxLayout()
        cfgSelLayout.addWidget(self._list)

        # Add add-config button
        t = translate("shell", "Add config ::: Add a new shell configuration")
        btn = QtWidgets.QPushButton(pyzo.icons.add, t)
        btn.setToolTip(t.tt)
        btn.clicked.connect(self._onBtnAddConfigPressed)
        cfgSelLayout.addWidget(btn)

        # Add duplicate button
        t = translate("shell", "Duplicate ::: Duplicate this shell configuration")
        btn = QtWidgets.QPushButton(pyzo.icons.page_white_copy, t)
        btn.setToolTip(t.tt)
        btn.clicked.connect(self._onBtnCopyConfigPressed)
        cfgSelLayout.addWidget(btn)

        # Add delete button
        t = translate("shell", "Delete ::: Delete this shell configuration")
        btn = QtWidgets.QPushButton(pyzo.icons.cancel, t)
        btn.setToolTip(t.tt)
        btn.clicked.connect(self._deleteConfig)
        cfgSelLayout.addWidget(btn)

        # Create dialog buttons
        cancelBut = QtWidgets.QPushButton("Cancel", self)
        okBut = QtWidgets.QPushButton("Done", self)
        cancelBut.clicked.connect(self.close)
        okBut.clicked.connect(self._applyAndClose)
        # Layout for dialog buttons
        buttonLayout = QtWidgets.QHBoxLayout()
        buttonLayout.addStretch(1)
        buttonLayout.addWidget(cancelBut)
        buttonLayout.addSpacing(10)
        buttonLayout.addWidget(okBut)

        # Layout the widgets
        configLayout = QtWidgets.QHBoxLayout()
        configLayout.addLayout(cfgSelLayout, 1)
        configLayout.addWidget(self._stack, 4)

        mainLayout = QtWidgets.QVBoxLayout(self)
        mainLayout.addSpacing(8)
        mainLayout.addLayout(configLayout)
        mainLayout.addLayout(buttonLayout, 0)
        self.setLayout(mainLayout)

        self.setMinimumSize(800, 600)

        # Add an entry if there's none
        if not pyzo.config.shellConfigs2:
            self._addConfig()

        # Fill config pages
        for item in pyzo.config.shellConfigs2:
            self._addConfig(item)

        self._list.setCurrentRow(0)

        self.show()

    def sizeHint(self):
        # We use this instead of self.resize(1024, 768) to avoid
        # a too large dialog window on smaller screens.
        return QtCore.QSize(1024, 768)

    def _addConfig(self, shellConfig=None, atIndex=-1, select=False):
        if atIndex == -1:
            atIndex = self._list.count()
        w = ShellInfoPage(self._stack)
        self._stack.addWidget(w)

        # QSignalBlocker has a context manager, but not in old PySide2
        try:
            blocker = QtCore.QSignalBlocker(self._list.model())
            self._list.insertItem(atIndex, w.listItem)
        finally:
            blocker.unblock()
        w.setInfo(shellConfig)
        if select:
            self._list.setCurrentItem(w.listItem)
            w.setFocus()

    def _getPageWidget(self, listItem):
        for i in range(self._stack.count()):
            w = self._stack.widget(i)
            if w.listItem is listItem:
                return w
        else:
            raise ValueError("item not found: " + repr(listItem))

    def _onListItemChanged(self, curItem, prevItem):
        if curItem is None:
            # last remaining item was deleted
            return
        w = self._getPageWidget(curItem)
        self._stack.setCurrentWidget(w)

    def _onBtnAddConfigPressed(self):
        self._addConfig(select=True)

    def _deleteConfig(self):
        w = self._stack.currentWidget()
        self._stack.removeWidget(w)
        self._list.takeItem(self._list.row(w.listItem))

        # make sure that there is at least one config
        if self._stack.count() == 0:
            self._addConfig(select=True)

    def _onBtnCopyConfigPressed(self):
        self._copyConfig()

    def _copyConfig(self, w0=None, index=None):
        if w0 is None:
            # Get original widget
            w0 = self._stack.currentWidget()
        # Build new info
        info = w0.getInfo().copy()
        info.name += " (2)"
        if index is None:
            index = self._list.row(w0.listItem) + 1
        self._addConfig(info, index, select=True)

    def _onDragDropRowCopied(self, indOriginalRow, indNewRow, later=False):
        if not later:
            # This is called when we are still in the "rowsInserted" event.
            # We need to let the insertion process finish before we modify anything.
            # Otherwise our change to the list entry text would be overwritten.
            pyzo.callLater(self._onDragDropRowCopied, indOriginalRow, indNewRow, True)
            return

        w0 = self._getPageWidget(self._list.item(indOriginalRow))

        # remove the item that was added via drag'n'drop copy and do our own copy
        self._list.takeItem(indNewRow)
        self._copyConfig(w0, indNewRow)

    def _applyAndClose(self, event=None):
        self._apply()
        self.close()

    def _apply(self):
        """Apply changes for all configs."""

        # Clear
        pyzo.config.shellConfigs2 = []

        # Set new versions. Note that although we recreate the list,
        # the list is filled with the orignal structs, so having a
        # reference to such a struct (as the shell has) will enable
        # you to keep track of any made changes.
        for i in range(self._list.count()):
            w = self._getPageWidget(self._list.item(i))
            pyzo.config.shellConfigs2.append(w.getInfo())



"""
Define tasks that can be executed by the file browser.
These inherit from proxies.Task and implement that specific interface.
"""

import re

from . import proxies


class SearchTask(proxies.Task):
    __slots__ = []

    def process(
        self,
        proxy,
        pattern=None,
        matchCase=False,
        regExp=False,
        wholeWords=False,
        excludeBinary=True,
        **rest,
    ):
        # Quick test
        if not pattern:
            return

        # Get text
        haystack = self._getText(proxy, excludeBinary)
        if not haystack:
            return

        flags = re.MULTILINE
        if not matchCase:
            flags |= re.IGNORECASE

        if not regExp:
            pattern = re.escape(pattern)

        if wholeWords:
            pattern = r"\b" + pattern + r"\b"

        indices = self._getIndicesRegExp(haystack, pattern, flags)

        # Return as lines
        if indices:
            return self._indicesToLines(haystack, indices)
        else:
            return []

    def _getText(self, proxy, excludeBinary=True):
        # Init
        path = proxy.path()
        fsProxy = proxy._fsProxy

        # # Get file size
        # try:
        #     size = fsProxy.fileSize(path) or 0
        # except NotImplementedError:
        #     size = 0

        # Always search Python files.
        if path.lower().endswith(".py"):
            excludeBinary = False

        if excludeBinary:
            # Check first few bytes for presence of zero bytes.
            # Git supposedly uses the same logic to determine whether
            # a file is binary. Note that files with certain encodings
            # other than utf-8 can be marked binary too.
            preamble_len = 8000
            bb_preamble = fsProxy.read(path, preamble_len)
            if bb_preamble is None:
                return None
            if 0 in bb_preamble:
                return None
            # Ready bytes (or re-use)
            if len(bb_preamble) < preamble_len:
                bb = bb_preamble
            else:
                bb = fsProxy.read(path)
        else:
            bb = fsProxy.read(path)

        # Convert to str
        if bb is None:
            return None
        else:
            return bb.decode("utf-8", errors="replace")

    def _getIndicesRegExp(self, text, pattern, flags):
        indices = []
        for match in re.finditer(pattern, text, flags):
            indices.append(match.start())
        return indices

    def _indicesToLines(self, text, indices):
        # Determine line endings
        LE = self._determineLineEnding(text)

        # Obtain line and line numbers
        lines = []
        for i in indices:
            # Get linenr and index of the line
            linenr = text.count(LE, 0, i) + 1
            i1 = text.rfind(LE, 0, i)
            i2 = text.find(LE, i)
            # Get line and strip
            if i1 < 0:
                i1 = 0
            line = text[i1:i2].strip()[:80]
            # Store
            lines.append((linenr, repr(line)))

        # Set result
        return lines

    def _determineLineEnding(self, text):
        """function to determine quickly whether LF or CR is used
        as line endings. Windows endings (CRLF) result in LF
        (you can split lines with either char).
        """
        i = 0
        LE = "\n"
        while i < len(text):
            i += 128
            LF = text.count("\n", 0, i)
            CR = text.count("\r", 0, i)
            if LF or CR:
                if CR > LF:
                    LE = "\r"
                break
        return LE


class PeekTask(proxies.Task):
    """To peek the high level structure of a task."""

    __slots__ = []

    stringStart = re.compile("(\"\"\"|'''|\"|')|#")
    endProgs = {
        "'": re.compile(r"(^|[^\\])(\\\\)*'"),
        '"': re.compile(r'(^|[^\\])(\\\\)*"'),
        "'''": re.compile(r"(^|[^\\])(\\\\)*'''"),
        '"""': re.compile(r'(^|[^\\])(\\\\)*"""'),
    }

    definition = re.compile(r"^(def|class)\s+(\w+)")

    def process(self, proxy):
        path = proxy.path()
        fsProxy = proxy._fsProxy

        # Search only Python files
        if not path.lower().endswith(".py"):
            return None

        # Get text
        bb = fsProxy.read(path)
        if bb is None:
            return
        try:
            text = bb.decode("utf-8")
            del bb
        except UnicodeDecodeError:
            # todo: right now we only do utf-8
            return

        # Parse
        return list(self._parseLines(text.splitlines()))

    def _parseLines(self, lines):
        stringEndProg = None

        linenr = 0
        for line in lines:
            linenr += 1

            # If we are in a triple-quoted multi-line string, find the end
            if stringEndProg is None:
                pos = 0
            else:
                endMatch = stringEndProg.search(line)
                if endMatch is None:
                    continue
                else:
                    pos = endMatch.end()
                    stringEndProg = None

            # Now process all tokens
            while True:
                match = self.stringStart.search(line, pos)

                if (
                    pos == 0
                ):  # If we are at the start of the line, see if we have a top-level class or method definition
                    end = len(line) if match is None else match.start()
                    definitionMatch = self.definition.search(line[:end])
                    if definitionMatch is not None:
                        if definitionMatch.group(1) == "def":
                            yield (linenr, "def " + definitionMatch.group(2))
                        else:
                            yield (linenr, "class " + definitionMatch.group(2))

                if match is None:
                    break  # Go to next line
                if match.group() == "#":
                    # comment
                    # yield 'C:'
                    break  # Go to next line
                else:
                    endMatch = self.endProgs[match.group()].search(line[match.end() :])
                    if endMatch is None:
                        if len(match.group()) == 3 or line.endswith("\\"):
                            # Multi-line string
                            stringEndProg = self.endProgs[match.group()]
                            break
                        else:  # incorrect end of single-quoted string
                            break

                    # yield 'S:' + (match.group() + line[match.end():][:endMatch.end()])
                    pos = match.end() + endMatch.end()


class DocstringTask(proxies.Task):
    __slots__ = []

    def process(self, proxy):
        path = proxy.path()
        fsProxy = proxy._fsProxy

        # Search only Python files
        if not path.lower().endswith(".py"):
            return None

        # Get text
        bb = fsProxy.read(path)
        if bb is None:
            return
        try:
            text = bb.decode("utf-8")
            del bb
        except UnicodeDecodeError:
            # todo: right now we only do utf-8
            return

        # Find docstring
        lines = []
        delim = None  # Not started, in progress, done
        count = 0
        for line in text.splitlines():
            count += 1
            if count > 200:
                break
            # Try to find a start
            if not delim:
                if line.startswith('"""'):
                    delim = '"""'
                    line = line.lstrip('"')
                elif line.startswith("'''"):
                    delim = "'''"
                    line = line.lstrip("'")
            # Try to find an end (may be on the same line as the start)
            if delim and delim in line:
                line = line.split(delim, 1)[0]
                count = 999999999  # Stop; we found the end
            # Add this line
            if delim:
                lines.append(line)

        # Limit number of lines
        if len(lines) > 16:
            lines = lines[:16] + ["..."]
        # Make text and strip
        doc = "\n".join(lines)
        doc = doc.strip()

        return doc


class RenameTask(proxies.Task):
    __slots__ = []

    def process(self, proxy, newpath=None, removeold=False):
        path = proxy.path()
        fsProxy = proxy._fsProxy

        if not newpath:
            return

        if removeold:
            # Works for files and dirs
            fsProxy.rename(path, newpath)
            # The fsProxy will detect that this file is now deleted
        else:
            # Work only for files: duplicate
            # Read bytes
            bb = fsProxy.read(path)
            if bb is None:
                return
            # write back with new name
            fsProxy.write(newpath, bb)


class CreateTask(proxies.Task):
    __slots__ = []

    def process(self, proxy, newpath=None, file=True):
        proxy.path()
        fsProxy = proxy._fsProxy

        if not newpath:
            return

        if file:
            fsProxy.write(newpath, b"")
        else:
            fsProxy.createDir(newpath)


class RemoveTask(proxies.Task):
    __slots__ = []

    def process(self, proxy):
        path = proxy.path()
        fsProxy = proxy._fsProxy

        # Remove
        fsProxy.remove(path)
        # The fsProxy will detect that this file is now deleted



import re
import sys
import os.path as op

import pyzo
from pyzo import translate
from pyzo.util import zon as ssdf

from . import QtCore, QtGui, QtWidgets
from . import proxies
from .tree import Tree
from .utils import cleanpath, isdir


class Browser(QtWidgets.QWidget):
    """A browser consists of an address bar, and tree view, and other
    widgets to help browse the file system. The browser object is responsible
    for tying the different browser-components together.

    It is also provides the API for dealing with starred dirs.
    """

    def __init__(self, parent, config, path=None):
        super().__init__(parent)

        # Store config
        self.config = config

        # Create star button
        self._projects = Projects(self)

        # Create path input/display lineEdit
        self._pathEdit = PathInput(self)

        # Create file system proxy
        self._fsProxy = proxies.NativeFSProxy()
        self.destroyed.connect(self._fsProxy.stop)

        # Create tree widget
        self._tree = Tree(self)
        self._tree.setPath(cleanpath(self.config.path))

        # Create name filter
        self._nameFilter = NameFilter(self)
        # self._nameFilter.lineEdit().setToolTip('File filter pattern')
        self._nameFilter.setToolTip(translate("filebrowser", "Filename filter"))
        self._nameFilter.setPlaceholderText(self._nameFilter.toolTip())

        # Create search filter
        self._searchFilter = SearchFilter(self)
        self._searchFilter.setToolTip(translate("filebrowser", "Search in files"))
        self._searchFilter.setPlaceholderText(self._searchFilter.toolTip())

        # Signals to sync path.
        # Widgets that can change the path transmit signal to _tree
        self._pathEdit.dirUp.connect(self._tree.setFocus)
        self._pathEdit.dirUp.connect(self._tree.setPathUp)
        self._pathEdit.dirChanged.connect(self._tree.setPath)
        self._projects.dirChanged.connect(self._tree.setPath)
        #
        self._nameFilter.filterChanged.connect(self._tree.onChanged)  # == update
        self._searchFilter.filterChanged.connect(self._tree.onChanged)
        # The tree transmits signals to widgets that need to know the path
        self._tree.dirChanged.connect(self._pathEdit.setPath)
        self._tree.dirChanged.connect(self._projects.setPath)

        self._layout()

        # Set and sync path ...
        if path is not None:
            self._tree.setPath(path)
        self._tree.dirChanged.emit(self._tree.path())

    def getImportWizard(self):
        # Lazy loading
        try:
            return self._importWizard
        except AttributeError:
            from .importwizard import ImportWizard

            self._importWizard = ImportWizard()

            return self._importWizard

    def _layout(self):
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        # layout.setSpacing(6)
        self.setLayout(layout)
        #
        layout.addWidget(self._projects)
        layout.addWidget(self._pathEdit)
        layout.addWidget(self._tree)
        #
        subLayout = QtWidgets.QHBoxLayout()
        subLayout.setSpacing(2)
        subLayout.addWidget(self._nameFilter, 5)
        subLayout.addWidget(self._searchFilter, 5)
        layout.addLayout(subLayout)

    def cleanUp(self):
        self._fsProxy.stop()

    def nameFilter(self):
        # return self._nameFilter.lineEdit().text()
        return self._nameFilter.text()

    def searchFilter(self):
        return {
            "pattern": self._searchFilter.text(),
            "matchCase": self.config.searchMatchCase,
            "regExp": self.config.searchRegExp,
            "wholeWords": self.config.wholeWords,
            "subDirs": self.config.searchSubDirs,
            "excludeBinary": self.config.searchExcludeBinary,
        }

    def setSearchText(self, needle, setFocus=False):
        """Set the text in the search field."""
        if self.config.searchRegExp:
            needle = re.escape(needle)
        self._searchFilter.setText(needle)
        if setFocus:
            self._searchFilter.setFocus()

    @property
    def expandedDirs(self):
        """The list of the expanded directories."""
        return self.parent().config.expandedDirs

    @property
    def starredDirs(self):
        """A list of the starred directories."""
        return [d.path for d in self.parent().config.starredDirs]

    def dictForStarredDir(self, path):
        """Return the dict of the starred dir corresponding to
        the given path, or None if no starred dir was found.
        """
        if not path:
            return None
        for d in self.parent().config.starredDirs:
            if op.normcase(d["path"]) == op.normcase(path):
                return d
        else:
            return None

    def addStarredDir(self, path):
        """Add the given path to the starred directories."""
        # Create new dict
        newProject = ssdf.new()
        newProject.path = op.normcase(path)  # Normalize case!
        newProject.name = op.basename(path)
        newProject.addToPythonpath = False
        # Add it to the config
        self.parent().config.starredDirs.append(newProject)
        # Update list
        self._projects.updateProjectList()

    def removeStarredDir(self, path):
        """Remove the given path from the starred directories.
        The path must exactlty match.
        """
        # Remove
        starredDirs = self.parent().config.starredDirs
        pathn = op.normcase(path)
        for d in starredDirs:
            if op.normcase(pathn) == op.normcase(d.path):
                starredDirs.remove(d)
        # Update list
        self._projects.updateProjectList()

    def test(self, sort=False):
        items = []
        for i in range(self._tree.topLevelItemCount()):
            item = self._tree.topLevelItem(i)
            items.append(item)
            # self._tree.removeItemWidget(item, 0)
        self._tree.clear()

        # items.sort(key=lambda x: x._path)
        for item in items[::-1]:
            self._tree.addTopLevelItem(item)

    def currentProject(self):
        """Return the ssdf dict for the current project, or None."""
        return self._projects.currentDict()


class LineEditWithToolButtons(QtWidgets.QLineEdit):
    """Line edit to which tool buttons (with icons) can be attached."""

    def __init__(self, parent):
        super().__init__(parent)
        self._leftButtons = []
        self._rightButtons = []

    def addButtonLeft(self, icon, willHaveMenu=False):
        return self._addButton(icon, willHaveMenu, self._leftButtons)

    def addButtonRight(self, icon, willHaveMenu=False):
        return self._addButton(icon, willHaveMenu, self._rightButtons)

    def _addButton(self, icon, willHaveMenu, L):
        # Create button
        button = QtWidgets.QToolButton(self)
        L.append(button)
        # Customize appearance
        button.setIcon(icon)
        button.setIconSize(QtCore.QSize(16, 16))
        button.setStyleSheet("QToolButton { border: none; padding: 0px; }")
        # button.setStyleSheet("QToolButton { border: none; padding: 0px; background-color:red;}");
        # Set behavior
        button.setCursor(QtCore.Qt.CursorShape.ArrowCursor)
        button.setPopupMode(button.ToolButtonPopupMode.InstantPopup)
        # Customize alignment
        if willHaveMenu:
            button.setToolButtonStyle(
                QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon
            )
            if sys.platform.startswith("win"):
                button.setText(" ")
        # Update self
        self._updateGeometry()
        return button

    def setButtonVisible(self, button, visible):
        for but in self._leftButtons:
            if but is button:
                but.setVisible(visible)
        for but in self._rightButtons:
            if but is button:
                but.setVisible(visible)
        self._updateGeometry()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._updateGeometry(True)

    def showEvent(self, event):
        super().showEvent(event)
        self._updateGeometry()

    def _updateGeometry(self, light=False):
        if not self.isVisible():
            return

        # Init
        rect = self.rect()

        # Determine padding and height
        paddingLeft, paddingRight, height = 1, 1, 0
        #
        for but in self._leftButtons:
            if but.isVisible():
                sz = but.sizeHint()
                height = max(height, sz.height())
                but.move(
                    int(1 + paddingLeft),
                    int(rect.bottom() + 1 - sz.height()) // 2,
                )
                paddingLeft += sz.width() + 1
        #
        for but in self._rightButtons:
            if but.isVisible():
                sz = but.sizeHint()
                paddingRight += sz.width() + 1
                height = max(height, sz.height())
                but.move(
                    int(rect.right() - 1 - paddingRight),
                    int(rect.bottom() + 1 - sz.height()) // 2,
                )

        # Set padding
        ss = "QLineEdit {{ padding-left: {}px; padding-right: {}px}} "
        self.setStyleSheet(ss.format(paddingLeft, paddingRight))

        # Set minimum size
        if not light:
            fw = QtWidgets.qApp.style().pixelMetric(
                QtWidgets.QStyle.PixelMetric.PM_DefaultFrameWidth
            )
            msz = self.minimumSizeHint()
            w = max(msz.width(), paddingLeft + paddingRight + 10)
            h = max(msz.height(), height + fw * 2 + 2)
            self.setMinimumSize(w, h)


class PathInput(LineEditWithToolButtons):
    """Line edit for selecting a path."""

    dirChanged = QtCore.Signal(
        str
    )  # Emitted when the user changes the path (and is valid)
    dirUp = QtCore.Signal()  # Emitted when user presses the up button

    def __init__(self, parent):
        super().__init__(parent)

        # Create up button
        self._upBut = self.addButtonLeft(pyzo.icons.folder_parent)
        self._upBut.clicked.connect(self.dirUp)

        # To receive focus events
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        # Set completion mode
        self.setCompleter(QtWidgets.QCompleter())
        c = self.completer()
        c.setCompletionMode(c.CompletionMode.InlineCompletion)

        # Set dir model to completer
        self._dirModel = QtWidgets.QFileSystemModel(c)
        self._dirModel.setFilter(
            QtCore.QDir.Filter.Dirs | QtCore.QDir.Filter.NoDotAndDotDot
        )
        # filter is not synchronized with NameFilter input (e.g. "!hidden")
        c.setModel(self._dirModel)
        if sys.platform != "win32":
            self._dirModel.setRootPath("/")
        # for win32 setRootPath is done self.setPath(...)

        # Connect signals
        # c.activated.connect(self.onActivated)
        self.textEdited.connect(self.onTextEdited)
        # self.textChanged.connect(self.onTextEdited)
        # self.cursorPositionChanged.connect(self.onTextEdited)

    def setPath(self, path):
        """Set the path to display. Does nothing if this widget has focus."""
        if sys.platform == "win32":
            oldDrive, _ = op.splitdrive(self.text())
            newDrive, _ = op.splitdrive(path)
            if oldDrive != newDrive:
                self._dirModel.setRootPath(newDrive)
        if not self.hasFocus():
            self.setText(path)
            self.checkValid()  # Reset style if it was invalid first

    def checkValid(self):
        # todo: This kind of violates the abstraction of the file system
        # ok for now, but we should find a different approach someday
        # Check
        text = self.text()
        dir = cleanpath(text)
        isvalid = text and isdir(dir) and op.isabs(dir)
        # Apply styling
        ss = self.styleSheet().replace("font-style:italic; ", "")
        if not isvalid:
            ss = ss.replace("QLineEdit {", "QLineEdit {font-style:italic; ")
        self.setStyleSheet(ss)
        # Return
        return isvalid

    def event(self, event):
        # Capture key events to explicitly apply the completion and
        # invoke checking whether the current text is a valid directory.
        # Test if QtGui is not None (can happen when reloading tools)
        if QtGui and isinstance(event, QtGui.QKeyEvent):
            k = QtCore.Qt.Key
            if event.key() in [k.Key_Tab, k.Key_Enter, k.Key_Return]:
                self.setText(self.text())  # Apply completion
                self.onTextEdited()  # Check if this is a valid dir
                return True
        return super().event(event)

    def onTextEdited(self, dummy=None):
        text = self.text()
        if self.checkValid():
            self.dirChanged.emit(cleanpath(text))

    def focusOutEvent(self, event=None):
        """On focusing out, make sure that the set path is correct."""
        if event is not None:
            super().focusOutEvent(event)

        path = self.parent()._tree.path()
        self.setPath(path)


class Projects(QtWidgets.QWidget):
    dirChanged = QtCore.Signal(str)  # Emitted when the user changes the project

    def __init__(self, parent):
        super().__init__(parent)

        # Init variables
        self._path = ""

        # Create combo button
        self._combo = QtWidgets.QComboBox(self)
        self._combo.setEditable(False)
        self.updateProjectList()

        # Create star button
        self._but = QtWidgets.QToolButton(self)
        self._but.setIcon(pyzo.icons.star3)
        self._but.setStyleSheet("QToolButton { padding: 0px; }")
        self._but.setIconSize(QtCore.QSize(18, 18))
        self._but.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self._but.setPopupMode(self._but.ToolButtonPopupMode.InstantPopup)
        #
        self._menu = QtWidgets.QMenu(self._but)
        self._menu.triggered.connect(self.onMenuTriggered)
        self.buildMenu()

        # Make equal height
        h = max(self._combo.sizeHint().height(), self._but.sizeHint().height())
        self._combo.setMinimumHeight(h)
        self._but.setMinimumHeight(h)

        # Connect signals
        self._but.pressed.connect(self.onButtonPressed)
        self._combo.activated.connect(self.onProjectSelect)

        # Layout
        layout = QtWidgets.QHBoxLayout(self)
        self.setLayout(layout)
        layout.addWidget(self._but)
        layout.addWidget(self._combo)
        layout.setSpacing(2)
        layout.setContentsMargins(0, 0, 0, 0)

    def currentDict(self):
        """Return the current project-dict, or None."""
        path = self._combo.itemData(self._combo.currentIndex())
        return self.parent().dictForStarredDir(path)

    def setPath(self, path):
        self._path = path
        # Find project index
        projectIndex, L = 0, 0
        pathn = op.normcase(path) + op.sep
        for i in range(self._combo.count()):
            projectPath = self._combo.itemData(i) + op.sep
            if pathn.startswith(projectPath) and len(projectPath) > L:
                projectIndex, L = i, len(projectPath)
        # Select project or not ...
        self._combo.setCurrentIndex(projectIndex)
        if projectIndex:
            self._but.setIcon(pyzo.icons.star2)
            self._but.setMenu(self._menu)
        else:
            self._but.setIcon(pyzo.icons.star3)
            self._but.setMenu(None)

    def updateProjectList(self):
        # Get sorted version of starredDirs
        starredDirs = self.parent().starredDirs
        starredDirs.sort(key=lambda p: self.parent().dictForStarredDir(p).name.lower())
        # Refill the combo box
        self._combo.clear()
        if starredDirs:
            self._combo.addItem(
                translate("filebrowser", "Projects:"), ""
            )  # No-project item
            for p in starredDirs:
                name = self.parent().dictForStarredDir(p).name
                self._combo.addItem(name, p)
        else:
            self._combo.addItem(
                translate("filebrowser", "Click star to bookmark current dir"), ""
            )

    def buildMenu(self):
        menu = self._menu
        menu.clear()

        # Add action to remove bookmark
        action = menu.addAction(translate("filebrowser", "Remove project"))
        action._id = "remove"
        action.setCheckable(False)

        # Add action to change name
        action = menu.addAction(translate("filebrowser", "Change project name"))
        action._id = "name"
        action.setCheckable(False)

        menu.addSeparator()

        # Add check action for adding to Pythonpath
        action = menu.addAction(translate("filebrowser", "Add path to Python path"))
        action._id = "pythonpath"
        action.setCheckable(True)
        d = self.currentDict()
        if d:
            checked = bool(d and d["addToPythonpath"])
            action.setChecked(checked)

        # Add action to cd to the project directory
        action = menu.addAction(
            translate("filebrowser", "Go to this directory in the current shell")
        )
        action._id = "cd"
        action.setCheckable(False)

    def onMenuTriggered(self, action):
        d = self.currentDict()
        if not d:
            return

        if action._id == "remove":
            # Remove this project
            self.parent().removeStarredDir(d.path)

        elif action._id == "name":
            # Open dialog to ask for name
            name, ok = QtWidgets.QInputDialog.getText(
                self.parent(),
                translate("filebrowser", "Project name"),
                translate("filebrowser", "New project name:"),
                text=d["name"],
            )
            if name and ok:
                d["name"] = name
            self.updateProjectList()

        elif action._id == "pythonpath":
            # Flip add-to-pythonpath flag
            d["addToPythonpath"] = not d["addToPythonpath"]

        elif action._id == "cd":
            # cd to the directory
            shell = pyzo.shells.getCurrentShell()
            if shell:
                shell.executeCommand("cd " + d.path + "\n")

    def onButtonPressed(self):
        if self._but.menu():
            # The directory is starred and has a menu. The user just
            # used the menu (or not). Update so it is up-to-date next time.
            self.buildMenu()
        else:
            # Not starred right now, create new project!
            self.parent().addStarredDir(self._path)
        # Update
        self.setPath(self._path)

    def onProjectSelect(self, index):
        path = self._combo.itemData(index)
        if path:
            # Go to dir
            self.dirChanged.emit(path)
        else:
            # Dummy item, reset
            self.setPath(self._path)


class NameFilter(LineEditWithToolButtons):
    """Combobox to filter by name."""

    filterChanged = QtCore.Signal()

    def __init__(self, parent):
        super().__init__(parent)

        # Create tool button, and attach the menu
        self._menuBut = self.addButtonRight(pyzo.icons["filter"], True)
        self._menu = QtWidgets.QMenu(self._menuBut)
        self._menu.triggered.connect(self.onMenuTriggered)
        self._menuBut.setMenu(self._menu)
        #
        # Add common patterns
        for pattern in [
            "*",
            "!hidden",
            "!*.pyc !hidden",
            "*.py *.pyw",
            "*.py *.pyw *.pyx *.pxd",
            "*.h *.c *.cpp",
        ]:
            self._menu.addAction(pattern)

        # Emit signal when value is changed
        self._lastValue = ""
        self.returnPressed.connect(self.checkFilterValue)
        self.editingFinished.connect(self.checkFilterValue)

        # Ensure the namefilter is in the config and initialize
        config = self.parent().config
        if "nameFilter" not in config:
            config.nameFilter = "!*.pyc"
        self.setText(config.nameFilter)

    def setText(self, value, test=False):
        """To initialize the name filter."""
        super().setText(value)
        if test:
            self.checkFilterValue()
        self._lastValue = value

    def checkFilterValue(self):
        value = self.text()
        if value != self._lastValue:
            self.parent().config.nameFilter = value
            self._lastValue = value
            self.filterChanged.emit()

    def onMenuTriggered(self, action):
        self.setText(action.text(), True)


class SearchFilter(LineEditWithToolButtons):
    """Line edit to do a search in the files."""

    filterChanged = QtCore.Signal()

    def __init__(self, parent):
        super().__init__(parent)

        # Create tool button, and attach the menu
        self._menuBut = self.addButtonRight(pyzo.icons["magnifier"], True)
        self._menu = QtWidgets.QMenu(self._menuBut)
        self._menu.triggered.connect(self.onMenuTriggered)
        self._menuBut.setMenu(self._menu)
        self.buildMenu()

        # Create cancel button
        self._cancelBut = self.addButtonRight(pyzo.icons["cancel"])
        self._cancelBut.setVisible(False)

        # Keep track of last value of search (initialized empty)
        self._lastValue = ""

        # Connect signals
        self._cancelBut.pressed.connect(self.onCancelPressed)
        self.textChanged.connect(self.updateCancelButton)
        self.editingFinished.connect(self.checkFilterValue)
        self.returnPressed.connect(self.forceFilterChanged)

    def onCancelPressed(self):
        """Clear text or build menu."""
        if self.text():
            super().clear()
            self.checkFilterValue()
        else:
            self.buildMenu()

    def checkFilterValue(self):
        value = self.text()
        if value != self._lastValue:
            self._lastValue = value
            self.filterChanged.emit()

    def forceFilterChanged(self):
        self._lastValue = self.text()
        self.filterChanged.emit()

    def updateCancelButton(self, text):
        visible = bool(self.text())
        self.setButtonVisible(self._cancelBut, visible)

    def buildMenu(self):
        config = self.parent().config
        menu = self._menu
        menu.clear()

        map = [
            ("searchMatchCase", False, translate("filebrowser", "Match case")),
            ("searchRegExp", False, translate("filebrowser", "RegExp")),
            ("wholeWords", False, translate("filebrowser", "Whole words")),
            ("searchSubDirs", True, translate("filebrowser", "Search in subdirs")),
            ("searchExcludeBinary", True, translate("filebrowser", "Exclude binary")),
        ]

        # Fill menu
        for option, default, description in map:
            if option is None:
                menu.addSeparator()
            else:
                # Make sure the option exists
                if option not in config:
                    config[option] = default
                # Make action in menu
                action = menu.addAction(description)
                action._option = option
                action.setCheckable(True)
                action.setChecked(bool(config[option]))

    def onMenuTriggered(self, action):
        config = self.parent().config
        option = action._option
        # Swap this option
        if option in config:
            config[option] = not config[option]
        else:
            config[option] = True
        # Update
        self.filterChanged.emit()



"""
The pyzokernel package contains the code for the Pyzo kernel process.
This kernel is designed to be relatively lightweight; i.e. most of
the work is done by the IDE.

See pyzokernel/start.py for more information.

"""


def printDirect(msg):
    """Small function that writes directly to the strm_out channel.
    This means that regardless if stdout was hijacked, the message ends
    up at the Pyzo shell. This keeps any hijacked stdout clean, and gets
    the message where you want it. In most cases this is just cosmetics:
    the Python banner is one example.
    """
    import sys

    sys._pyzoInterpreter.context._strm_out.send(msg)



"""
This file provides the pyzo history viewer. It contains two main components: the
History class, which is a Qt model, and the PyzoHistoryViewer, which is a Qt view



"""

import types
import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets  # noqa
from pyzo import translate
from pyzo.core.menu import Menu

tool_name = translate("pyzoHistoryViewer", "History viewer")
tool_summary = "Shows the last used commands."


def copyForListWidget(self):
    self.parent().copy()  # call the copy method of class PyzoHistoryViewer


class PyzoHistoryViewer(QtWidgets.QWidget):
    """
    The history viewer has several ways of using the data stored in the history:
     - double click a single item to execute in the current shell
     - drag and drop one or multiple selected lines into the editor or any
       other widget or application accepting plain text
     - copy selected items using the copy item in the pyzo edit menu
     - copy selected items using the context menu
     - execute selected items in the current shell using the context menu
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Widgets
        self._search = QtWidgets.QLineEdit(self)
        self._list = QtWidgets.QListWidget(self)

        # Set monospace
        font = self._list.font()
        font.setFamily(pyzo.config.view.fontname)
        self._list.setFont(font)

        # Layout
        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)
        layout.addWidget(self._search, 0)
        layout.addWidget(self._list, 1)
        # set margins
        margin = pyzo.config.view.widgetMargin
        layout.setContentsMargins(margin, margin, margin, margin)

        # Customize line edit
        self._search.setPlaceholderText(translate("menu", "Search"))
        self._search.textChanged.connect(self._on_search)

        # Drag/drop
        self._list.setSelectionMode(self._list.SelectionMode.ExtendedSelection)
        self._list.setDragEnabled(True)
        self._list.doubleClicked.connect(self._onDoubleClicked)

        # Context menu
        self._menu = Menu(self, translate("menu", "History"))
        self._menu.addItem(
            translate("menu", "Copy ::: Copy selected lines"),
            pyzo.icons.page_white_copy,
            self.copy,
            "copy",
        )
        self._menu.addItem(
            translate("menu", "Run ::: Run selected lines in current shell"),
            pyzo.icons.run_lines,
            self.runSelection,
            "run",
        )
        self._menu.addItem(
            translate("menu", "Remove ::: Remove selected history items(s)"),
            pyzo.icons.delete,
            self.removeSelection,
            "remove",
        )

        self._list.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self._list.customContextMenuRequested.connect(
            self._onCustomContextMenuRequested
        )

        # Populate
        for command in pyzo.command_history.get_commands():
            self._list.addItem(command)

        # Scroll to end of list on start up
        self._list.setCurrentRow(self._list.count() - 1)
        item = self._list.currentItem()
        self._list.scrollToItem(item)

        # Keep up to date ...
        pyzo.command_history.command_added.connect(self._on_command_added)
        pyzo.command_history.command_removed.connect(self._on_command_removed)
        pyzo.command_history.commands_reset.connect(self._on_commands_reset)

        # Patch a copy method to the list widget. We need this because the Ctrl+C
        # shortcut triggers the "Edit -> Copy" menu action. And in _editItemCallback
        # the "copy" method of the widget with the focus is called.
        # Otherwise Ctrl+C does not work to copy selected lines in the widget.
        self._list.copy = types.MethodType(copyForListWidget, self._list)

    def _on_search(self):
        needle = self._search.text()
        for i in range(self._list.count()):
            item = self._list.item(i)
            item.setHidden(bool(needle and needle not in item.text()))

    ## Keep track of history

    def _on_command_added(self, command):
        item = QtWidgets.QListWidgetItem(command, self._list)
        self._list.addItem(item)
        needle = self._search.text()
        item.setHidden(bool(needle and needle not in command))
        self._list.scrollToItem(item)

    def _on_command_removed(self, index):
        self._list.takeItem(index)

    def _on_commands_reset(self):
        self._list.clear()
        for command in pyzo.command_history.get_commands():
            self._list.addItem(command)

    ## User actions

    def _onCustomContextMenuRequested(self, pos):
        self._menu.popup(self._list.viewport().mapToGlobal(pos))

    def copy(self, event=None):
        text = "\n".join(i.text() for i in self._list.selectedItems())
        QtWidgets.qApp.clipboard().setText(text)

    def removeSelection(self, event=None):
        indices = [i.row() for i in self._list.selectedIndexes()]
        for i in reversed(sorted(indices)):
            pyzo.command_history.pop(i)

    def runSelection(self, event=None):
        commands = [i.text() for i in self._list.selectedItems()]
        shell = pyzo.shells.getCurrentShell()
        if shell is not None:
            for command in reversed(commands):
                pyzo.command_history.append(command)
            shell.executeCommand("\n".join(commands) + "\n")
            if len(commands) > 1 and commands[-1].startswith(" "):
                shell.executeCommand("\n")  # finalize multi-command

    def _onDoubleClicked(self, index):
        text = "\n".join(i.text() for i in self._list.selectedItems())
        shell = pyzo.shells.getCurrentShell()
        if shell is not None:
            shell.executeCommand(text + "\n")
            # Do not update history? Was this intended?


if __name__ == "__main__":
    import pyzo.core.main

    m = pyzo.core.main.MainWindow()
    view = PyzoHistoryViewer()
    view.show()



import importlib as _importlib


_QT_WRAPPERS = ("PySide6", "PyQt6", "PySide2", "PyQt5")


def _get_desired_api():
    import os

    desired_api = None
    s = os.environ.get("QT_API", "").strip().lower()
    if len(s) > 0:
        for module_name in _QT_WRAPPERS:
            if module_name.lower() == s:
                desired_api = module_name
                break
    return desired_api


def _load_modules(desired_api):
    import sys

    loaded_modules = [s for s in sys.modules if s in _QT_WRAPPERS]
    if desired_api in loaded_modules:
        # use the preferred qt wrapper that is already loaded
        return desired_api
    if len(loaded_modules) > 0:
        # use another qt wrapper that is already loaded
        api = loaded_modules[0]
        print('using Qt binding "{}" because it is already loaded'.format(api))
        return api

    # no qt modules loaded, try to import the preferred one first
    if desired_api is not None:
        order_to_try = [desired_api] + [s for s in _QT_WRAPPERS if s != desired_api]
    else:
        order_to_try = list(_QT_WRAPPERS)

    for api in order_to_try:
        try:
            _importlib.import_module(api)
        except ModuleNotFoundError:
            pass
        else:
            if desired_api is not None and api != desired_api:
                print('using Qt binding "{}" instead of "{}"'.format(api, desired_api))
            return api
    raise ImportError("no suitable Qt wrapper found")


API = _load_modules(_get_desired_api())


def _get_versions():
    """returns the qt and wrapper version as strings"""
    mod = _importlib.import_module(API)
    _importlib.import_module(API + ".QtCore")
    if API.startswith("PySide"):
        qt_version = mod.QtCore.__version__
        wrapper_version = mod.__version__
    else:
        qt_version = mod.QtCore.QT_VERSION_STR
        wrapper_version = mod.QtCore.PYQT_VERSION_STR
    return qt_version, wrapper_version


QT_VERSION_STR, QT_WRAPPER_VERSION_STR = _get_versions()

from . import qtutils as qtutils  # we have to write it like this to satisfy ruff

del _get_versions, _load_modules, _get_desired_api


## Qt components

QtCore = _importlib.import_module(API + ".QtCore")
QtGui = _importlib.import_module(API + ".QtGui")
QtWidgets = _importlib.import_module(API + ".QtWidgets")
QtPrintSupport = _importlib.import_module(API + ".QtPrintSupport")


## QtCore fixes

if API in ("PyQt5", "PyQt6"):
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.SignalInstance = QtCore.pyqtBoundSignal
    QtCore.Slot = QtCore.pyqtSlot
    QtCore.Property = QtCore.pyqtProperty

if API in ("PySide2", "PyQt5"):
    QtCore.QLibraryInfo.path = QtCore.QLibraryInfo.location
    QtCore.QLibraryInfo.LibraryPath = QtCore.QLibraryInfo

## QtGui fixes

if API in ("PySide2", "PyQt5"):
    QtGui.QAction = QtWidgets.QAction
    QtGui.QActionGroup = QtWidgets.QActionGroup

    _QPointF = QtCore.QPointF
    QtGui.QMouseEvent.position = lambda self: _QPointF(self.pos())
    QtGui.QDropEvent.position = lambda self: _QPointF(self.posF())

    # For Qt5 we need something like QFontDatabase = QFontDatabase()
    # but with instance creation just when first accessing the class.
    def _getFontDatabase(name):
        if not hasattr(_QFontDatabaseWrapper, "_instance"):
            _QFontDatabaseWrapper._instance = _QFontDatabaseWrapper._original()
        return getattr(_QFontDatabaseWrapper._instance, name)

    class _QFontDatabaseWrapper:
        _original = QtGui.QFontDatabase

        def __getattribute__(self, name):
            return _getFontDatabase(name)

    QtGui.QFontDatabase = _QFontDatabaseWrapper()

## QtWidgets fixes

if API == "PyQt6":
    QtWidgets.QFileSystemModel = QtGui.QFileSystemModel


@property
def old_exec(self):
    return self.exec_


@property
def old_print(self):
    return self.print_


if API in ("PySide2", "PyQt5"):
    QtWidgets.QMenu.exec = old_exec
    QtWidgets.QApplication.exec = old_exec
    QtWidgets.QPlainTextEdit.print = old_print

if API == "PySide6":
    # In PySide6 v6.10.1 QPlainTextEdit still has "print_" but no "print" method.
    # We check that before in case that it will be added in a future version.
    if not hasattr(QtWidgets.QPlainTextEdit, "print"):
        QtWidgets.QPlainTextEdit.print = old_print

del old_exec
del old_print



"""Module statusbar

Functionality for status bar in pyzo.

"""

from pyzo.qt import QtWidgets


class StatusBar(QtWidgets.QStatusBar):
    """
    Add a statusbar to main window

    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Cursor position
        self.cursor_pos = QtWidgets.QLabel(self)
        self.addPermanentWidget(self.cursor_pos)

    def updateCursorInfo(self, editor):
        # Get current line number
        nrow = 0
        ncol = 0
        sel = 0
        if editor:
            cur = editor.textCursor()
            nrow = cur.blockNumber() + 1
            ncol = cur.positionInBlock() + 1
            sel = cur.selectionEnd() - cur.selectionStart()

        position_txt = "Line: {}, Column: {}, Sel.: {} ".format(nrow, ncol, sel)
        self.cursor_pos.setText(position_txt)



from . import API

if API in ("PySide2", "PySide6"):
    if API == "PySide2":
        import shiboken2 as _shiboken
    else:
        import shiboken6 as _shiboken

    def isDeleted(qtObj):
        return not _shiboken.isValid(qtObj)

    def isOwnedByPython(qtObj):
        try:
            # for shiboken6 >= 6.8.0 and shiboken2 5.15.2.1
            return _shiboken.ownedByPython(qtObj)
        except AttributeError:
            pass
        return _shiboken.isOwnedByPython(qtObj)

elif API in ("PyQt5", "PyQt6"):
    if API == "PyQt5":
        import PyQt5.sip as _sip
    else:
        import PyQt6.sip as _sip

    def isDeleted(qtObj):
        return _sip.isdeleted(qtObj)

    def isOwnedByPython(qtObj):
        return _sip.ispyowned(qtObj)



# This code is loosely based on the event system of Visvis and on the
# signals system of Qt.

# Note: Python has a built-in module (sched) that does some of the things
# here. However, only since Python3.3 is this built-in functionality
# thread safe. And we need thread safety!

"""Module yoton.events

Yoton comes with a simple event system to enable event-driven applications.

All channels are capable of running without the event system, but some
channels have limitations. See the documentation of the channels for
more information. Note that signals only work if events are processed.

"""

import time
import threading
import weakref

from yoton.misc import getErrorMsg, PackageQueue


class CallableObject(object):
    """CallableObject(callable)

    A class to hold a callable. If it is a plain function, its reference
    is held (because it might be a closure). If it is a method, we keep
    the function name and a weak reference to the object. In this way,
    having for instance a signal bound to a method, the object is not
    prevented from being cleaned up.

    """

    __slots__ = ["_func", "_ob"]  # Use __slots__ to reduce memory footprint

    def __init__(self, c):
        # Check
        if not callable(c):
            raise ValueError("Error: given callback is not callable.")

        # Store function and object
        if hasattr(c, "__self__"):
            # Method, store object and method name
            self._ob = weakref.ref(c.__self__)
            self._func = c.__func__.__name__
        elif hasattr(c, "im_self"):
            # Method in older Python
            self._ob = weakref.ref(c.im_self)
            self._func = c.im_func.__name__
        else:
            # Plain function
            self._func = c
            self._ob = None

    def isdead(self):
        """Get whether the weak ref is dead."""
        if self._ob:
            # Method
            return self._ob() is None
        else:
            return False

    def compare(self, other):
        """compare this instance with another."""
        if self._ob and other._ob:
            return (self._ob() is other._ob()) and (self._func == other._func)
        elif not (self._ob or other._ob):
            return self._func == other._func
        else:
            return False

    def __str__(self):
        return self._func.__str__()

    def call(self, *args, **kwargs):
        """call(*args, **kwargs)
        Call the callable. Exceptions are caught and printed.
        """
        if self.isdead():
            return

        # Get function
        try:
            if self._ob:
                func = getattr(self._ob(), self._func)
            else:
                func = self._func
        except Exception:
            return

        # Call it
        try:
            return func(*args, **kwargs)
        except Exception:
            funcname = getattr(self._func, "__name__", self._func)
            print("Exception while handling event (in %s):" % (funcname,))
            print(getErrorMsg())


class Event(object):
    """Event(callable, *args, **kwargs)

    An Event instance represents something that is going to be done.
    It consists of a callable and arguments to call it with.

    Instances of this class populate the event queue.

    """

    __slots__ = ["_args", "_callable", "_kwargs", "_timeout"]

    def __init__(self, callable, *args, **kwargs):
        if isinstance(callable, CallableObject):
            self._callable = callable
        else:
            self._callable = CallableObject(callable)
        self._args = args
        self._kwargs = kwargs

    def dispatch(self):
        """dispatch()
        Call the callable with the arguments and keyword-arguments specified
        at initialization.
        """
        self._callable.call(*self._args, **self._kwargs)

    def _on_timeout(self):
        """This is what theTimerThread calls."""
        app.post_event(self)


class Signal:
    """Signal()

    The purpose of a signal is to provide an interface to bind/unbind
    to events and to fire them.

    One can bind() or unbind() a callable to the signal. When emitted, an
    event is created for each bound handler. Therefore, the event loop
    must run for signals to work.

    Some signals call the handlers using additional arguments to
    specify specific information.

    """

    def __init__(self):
        self._handlers = []

    @property
    def type(self):
        """The type (__class__) of this event."""
        return self.__class__

    def bind(self, func):
        """bind(func)

        Add an eventhandler to this event.

        The callback/handler (func) must be a callable. It is called
        with one argument: the event instance, which can contain
        additional information about the event.

        """

        # make callable object (checks whether func is callable)
        cnew = CallableObject(func)

        # check -> warn
        for c in self._handlers:
            if cnew.compare(c):
                print("Warning: handler %s already present for %s" % (func, self))
                return

        # add the handler
        self._handlers.append(cnew)

    def unbind(self, func=None):
        """unbind(func=None)

        Unsubscribe a handler, If func is None, remove all handlers.

        """
        if func is None:
            self._handlers[:] = []
        else:
            cref = CallableObject(func)
            for c in self._handlers[:]:
                # remove if callable matches func or object is destroyed
                if c.compare(cref) or c.isdead():
                    self._handlers.remove(c)

    def emit(self, *args, **kwargs):
        """emit(*args, **kwargs)

        Emit the signal, calling all bound callbacks with *args and **kwargs.
        An event is queued for each callback registered to this signal.
        Therefore it is safe to call this method from another thread.

        """

        # Add an event for each callback
        toremove = []
        for func in self._handlers:
            if func.isdead():
                toremove.append(func)
            else:
                event = Event(func, *args, **kwargs)
                app.post_event(event)

        # Remove dead ones
        for func in toremove:
            self._handlers.remove(func)

    def emit_now(self, *args, **kwargs):
        """emit_now(*args, **kwargs)

        Emit the signal *now*. All handlers are called from the calling
        thread. Beware, this should only be done from the same thread
        that runs the event loop.

        """

        # Add an event for each callback
        toremove = []
        for func in self._handlers:
            if func.isdead():
                toremove.append(func)
            else:
                func.call(*args, **kwargs)

        # Remove dead ones
        for func in toremove:
            self._handlers.remove(func)


class TheTimerThread(threading.Thread):
    """TheTimerThread is a singleton thread that is used by all timers
    and delayed events to wait for a while (in a separate thread) and then
    post an event to the event-queue. By sharing a single thread timers
    stay lightweight and there is no time spent on initializing or tearing
    down threads. The downside is that when there are a lot of timers running
    at the same time, adding a timer may become a bit inefficient because
    the registered objects must be sorted each time an object is added.
    """

    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True
        self._exit = False
        self._timers = []
        self._somethingChanged = False
        self._condition = threading.Condition(threading.Lock())

    def stop(self, timeout=1.0):
        self._exit = True
        with self._condition:
            self._condition.notify()
        self.join(timeout)

    def add(self, timer):
        """add(timer)
        Add item to the list of objects to track. The object should
        have a _timeout attribute, representing the time.time() at which
        it runs out, and an _on_timeout() method to call when it does.
        """
        # Check
        if not (hasattr(timer, "_timeout") and hasattr(timer, "_on_timeout")):
            raise ValueError("Cannot add this object to theTimerThread.")
        # Add item
        with self._condition:
            if timer not in self._timers:
                self._timers.append(timer)
                self._sort()
                self._somethingChanged = True
            self._condition.notify()

    def _sort(self):
        self._timers.sort(key=lambda x: x._timeout, reverse=True)

    def discard(self, timer):
        """Stop the timer if it hasn't finished yet"""
        with self._condition:
            if timer in self._timers:
                self._timers.remove(timer)
            self._somethingChanged = True
            self._condition.notify()

    def run(self):
        with self._condition:
            self._mainloop()

    def _mainloop(self):
        while not self._exit:
            # Set flag
            self._somethingChanged = False

            # Wait here, in wait() the underlying lock is released
            if self._timers:
                timer = self._timers[-1]
                timeout = timer._timeout - time.time()
                if timeout > 0:
                    self._condition.wait(timeout)
            else:
                timer = None
                self._condition.wait()

            # Here the lock has been re-acquired. Take action?
            if self._exit:
                break
            if (timer is not None) and (not self._somethingChanged):
                if timer._on_timeout():
                    self._sort()  # Keep and resort
                else:
                    self._timers.pop()  # Pop


# Instantiate and start the single timer thread
# We can do this as long as we do not wait for the thread, and the thread
# does not do any imports:
# http://docs.python.org/library/threading.html#importing-in-threaded-code
theTimerThread = TheTimerThread()
theTimerThread.start()


class Timer(Signal):
    """Timer(interval=1.0, oneshot=True)

    Timer class. You can bind callbacks to the timer. The timer is
    fired when it runs out of time.

    Parameters
    ----------
    interval : number
        The interval of the timer in seconds.
    oneshot : bool
        Whether the timer should do a single shot, or run continuously.

    """

    def __init__(self, interval=1.0, oneshot=True):
        Signal.__init__(self)

        # store Timer specific properties
        self.interval = interval
        self.oneshot = oneshot
        #
        self._timeout = 0

    @property
    def interval(self):
        """Gets the timer's interval in seconds."""
        return self._interval

    @interval.setter
    def interval(self, timer_interval_secs):
        """Sets the timer's interval in seconds."""
        if not isinstance(timer_interval_secs, (int, float)):
            raise ValueError("interval must be a float or integer.")
        if timer_interval_secs <= 0:
            raise ValueError("interval must be larger than 0.")
        self._interval = float(timer_interval_secs)

    @property
    def oneshot(self):
        """returns whether this is a oneshot timer. If False, it runs continuously."""
        return self._oneshot

    @oneshot.setter
    def oneshot(self, oneshot):
        """sets whether this is a oneshot timer. If False, it runs continuously."""
        self._oneshot = bool(oneshot)

    @property
    def running(self):
        """Get whether the timer is running."""
        return self._timeout > 0

    def start(self, interval=None, oneshot=None):
        """start(interval=None, oneshot=None)

        Start the timer. If interval or oneshot are not given,
        their current values are used.

        """
        # set properties?
        if interval is not None:
            self.interval = interval
        if oneshot is not None:
            self.oneshot = oneshot

        # put on
        self._timeout = time.time() + self.interval
        theTimerThread.add(self)

    def stop(self):
        """stop()

        Stop the timer from running.

        """
        theTimerThread.discard(self)
        self._timeout = 0

    def _on_timeout(self):
        """Method to call when the timer finishes. Called from
        event-loop-thread.
        """

        # Emit signal
        self.emit()
        # print('timer timeout', self.oneshot)
        # Do we need to stop it now, or restart it
        if self.oneshot:
            # This timer instance is removed from the list of Timers
            # when the timeout is reached.
            self._timeout = 0
            return False
        else:
            # keep in the thread
            self._timeout = time.time() + self.interval
            return True


class YotonApplication(object):
    """YotonApplication

    Represents the yoton application and contains functions for
    the event system. Multiple instances can be created, they will
    all operate on the same event queue and share attributes
    (because these are on the class, not on the instance).

    One instance of this class is always accessible via yoton.app.
    For convenience, several of its methods are also accessible
    directly from the yoton module namespace.

    """

    # Event queues
    _event_queue = PackageQueue(10000, "new")

    # Flag to stop event loop
    _stop_event_loop = False

    # Flag to signal whether we are in an event loop
    # Can be set externally if the event loop is hijacked.
    _in_event_loop = False

    # To allow other event loops to embed the yoton event loop
    _embedding_callback1 = None  # The reference
    _embedding_callback2 = None  # Used in post_event

    def call_later(self, func, timeout=0.0, *args, **kwargs):
        """call_later(func, timeout=0.0, *args, **kwargs)

        Call the given function after the specified timeout.

        Parameters
        ----------
        func : callable
            The function to call.
        timeout : number
            The time to wait in seconds. If zero, the event is put on the event
            queue. If negative, the event will be put at the front of the event
            queue, so that it's processed asap.
        args : arguments
            The arguments to call func with.
        kwargs: keyword arguments.
            The keyword arguments to call func with.

        """

        # Wrap the object in an event
        event = Event(func, *args, **kwargs)

        # Put it in the queue
        if timeout > 0:
            self.post_event_later(event, timeout)
        elif timeout < 0:
            self.post_event_asap(event)  # priority event
        else:
            self.post_event(event)

    def post_event(self, event):
        """post_event(events)

        Post an event to the event queue.

        """
        YotonApplication._event_queue.push(event)
        #
        if YotonApplication._embedding_callback2 is not None:
            YotonApplication._embedding_callback2 = None
            YotonApplication._embedding_callback1()

    def post_event_asap(self, event):
        """post_event_asap(event)

        Post an event to the event queue. Handle as soon as possible;
        putting it in front of the queue.

        """
        YotonApplication._event_queue.insert(event)
        #
        if YotonApplication._embedding_callback2 is not None:
            YotonApplication._embedding_callback2 = None
            YotonApplication._embedding_callback1()

    def post_event_later(self, event, delay):
        """post_event_later(event, delay)

        Post an event to the event queue, but with a certain delay.

        """
        event._timeout = time.time() + delay
        theTimerThread.add(event)
        # Calls post_event in due time

    def process_events(self, block=False):
        """process_events(block=False)

        Process all yoton events currently in the queue.
        This function should be called periodically
        in order to keep the yoton event system running.

        block can be False (no blocking), True (block), or a float
        blocking for maximally 'block' seconds.

        """
        # Reset callback for the embedding event loop
        YotonApplication._embedding_callback2 = YotonApplication._embedding_callback1

        # Process events
        try:
            while True:
                event = YotonApplication._event_queue.pop(block)
                event.dispatch()
                block = False  # Proceed until there are now more events
        except PackageQueue.Empty:
            pass

    def start_event_loop(self):
        """start_event_loop()

        Enter an event loop that keeps calling yoton.process_events().
        The event loop can be stopped using stop_event_loop().

        """

        # Don't go if we are in an event loop
        if YotonApplication._in_event_loop:
            return

        # Set flags
        YotonApplication._stop_event_loop = False
        YotonApplication._in_event_loop = True

        try:
            # Keep blocking for 3 seconds so a keyboardinterrupt still works
            while not YotonApplication._stop_event_loop:
                self.process_events(3.0)
        finally:
            # Unset flag
            YotonApplication._in_event_loop = False

    def stop_event_loop(self):
        """stop_event_loop()

        Stops the event loop if it is running.

        """
        if not YotonApplication._stop_event_loop:
            # Signal stop
            YotonApplication._stop_event_loop = True
            # Push an event so that process_events() unblocks

            def dummy():
                pass

            self.post_event(Event(dummy))

    def embed_event_loop(self, callback):
        """embed_event_loop(callback)

        Embed the yoton event loop in another event loop. The given callback
        is called whenever a new yoton event is created. The callback
        should create an event in the other event-loop, which should
        lead to a call to the process_events() method. The given callback
        should be thread safe.

        Use None as an argument to disable the embedding.

        """
        YotonApplication._embedding_callback1 = callback
        YotonApplication._embedding_callback2 = callback


# Instantiate an application object
app = YotonApplication()



import re

import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets

tool_name = pyzo.translate("pyzoWorkspace", "Workspace")
tool_summary = pyzo.translate(
    "pyzoWorkspace", "Lists the variables in the current shell's namespace."
)


def splitName(name):
    """Split an object name in parts, taking dots and indexing into account."""
    parts = name.replace("[", ".[").split(".")
    return [p for p in parts if p]


def joinName(parts):
    """Join the parts of an object name, taking dots and indexing into account."""
    return ".".join(parts).replace(".[", "[")


def wildcardsToRegExp(searchText):
    """converts a search text with wildcards to a regular expression string

    e.g.: 'ab)c*de?f??gh*' --> 'ab\\)c.*?de.f..gh.*?'
    """
    ll = []
    for s in re.split(r"(\*|\?)", searchText):
        if s == "*":
            ll.append(".*?")
        elif s == "?":
            ll.append(".")
        else:
            ll.append(re.escape(s))
    return "".join(ll)


class WorkspaceProxy(QtCore.QObject):
    """WorkspaceProxy

    A proxy class to handle the asynchonous behaviour of getting information
    from the shell. The workspace tool asks for a certain name, and this
    class notifies when new data is available using a qt signal.

    """

    haveNewData = QtCore.Signal()

    def __init__(self):
        super().__init__()

        # Variables
        self._variables = []

        # Element to get more info of
        self._name = ""

        # Bind to events
        pyzo.shells.currentShellChanged.connect(self._onCurrentShellChanged)
        pyzo.shells.currentShellStateChanged.connect(self._onCurrentShellStateChanged)

        # Initialize
        self._onCurrentShellStateChanged()

    def addNamePart(self, part):
        """Add a part to the name."""
        parts = splitName(self._name)
        parts.append(part)
        self.setName(joinName(parts))

    def setName(self, name):
        """Set the name that we want to know more of."""
        self._name = name

        shell = pyzo.shells.getCurrentShell()
        if shell:
            future = shell._request.dir2(self._name)
            future.add_done_callback(self._processResponse)

    def goUp(self):
        """Cut the last part off the name."""
        parts = splitName(self._name)
        if parts:
            parts.pop()
        self.setName(joinName(parts))

    def _onCurrentShellChanged(self):
        """When no shell is selected now, update this. In all other cases,
        the _onCurrentShellStateChanged will be fired too.
        """
        shell = pyzo.shells.getCurrentShell()
        if not shell:
            self._variables = []
            self.haveNewData.emit()

    def _onCurrentShellStateChanged(self):
        """Do a request for information!"""
        self.updateVariables()

    def updateVariables(self):
        shell = pyzo.shells.getCurrentShell()
        if not shell:
            self._variables = []
        elif shell._state not in ("Busy", "Very busy"):
            future = shell._request.dir2(self._name)
            future.add_done_callback(self._processResponse)

    def _processResponse(self, future):
        """We got a response, update our list and notify the tree."""

        response = []

        # Process future
        if future.cancelled():
            pass  # print('Introspect cancelled') # No living kernel
        elif future.exception():
            print("Introspect-queryDoc-exception: ", future.exception())
        else:
            response = future.result()

        self._variables = response
        self.haveNewData.emit()


class WorkspaceItem(QtWidgets.QTreeWidgetItem):
    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            return float(self.text(column).strip("[]")) > float(
                otherItem.text(column).strip("[]")
            )
        except ValueError:
            return self.text(column) > otherItem.text(column)


class WorkspaceTree(QtWidgets.QTreeWidget):
    """WorkspaceTree

    The tree that displays the items in the current namespace.
    I first thought about implementing this using the mode/view
    framework, but it is so much work and I can't seem to fully
    understand how it works :(

    The QTreeWidget is so very simple and enables sorting very
    easily, so I'll stick with that ...

    """

    def __init__(self, parent):
        super().__init__(parent)

        self._config = parent._config
        self._compiledRegExp = re.compile(r".*")

        # Set header stuff
        self.setHeaderHidden(False)
        self.setColumnCount(3)
        self.setHeaderLabels(
            [
                pyzo.translate("pyzoWorkspace", "Name"),
                pyzo.translate("pyzoWorkspace", "Type"),
                pyzo.translate("pyzoWorkspace", "Repr"),
            ]
        )
        # self.setColumnWidth(0, 100)
        self.setSortingEnabled(True)

        # Nice rows
        self.setAlternatingRowColors(True)
        self.setRootIsDecorated(False)

        # Create proxy
        self._proxy = WorkspaceProxy()
        self._proxy.haveNewData.connect(self.fillWorkspace)

        # For menu
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.DefaultContextMenu)
        self._menu = QtWidgets.QMenu()
        self._menu.triggered.connect(self._contextMenuTriggered)

        # Bind to events
        self.itemActivated.connect(self.onItemExpand)

        self._startUpVariables = frozenset(("In", "Out", "exit", "get_ipython", "quit"))

        self.setToolTip(
            "keyboard shortcuts for selected variable:\n"
            "h ... show variable info in Interactive help tool\n"
            "v ... add variable to Expression viewer tool\n"
            "p ... print(variable)\n"
            "r ... print(repr(variable))\n"
            "RETURN/ENTER ... show namespace\n"  # this is done via signal itemActivated
            "BACKSPACE ... switch to parent namespace\n"
            "DEL ... del variable"
        )

    def keyPressEvent(self, event):
        key = event.key()
        item = self.currentItem()
        if item is not None and not event.modifiers():
            if key == QtCore.Qt.Key.Key_Delete:
                self._performCommand("Delete", item)
                return
            if key == QtCore.Qt.Key.Key_H:
                self._performCommand("Show help", item)
                return
            if key == QtCore.Qt.Key.Key_V:
                self._performCommand("Add to Expr Viewer", item)
                return
            if key == QtCore.Qt.Key.Key_P:
                self._performCommand("str", item)
                return
            if key == QtCore.Qt.Key.Key_R:
                self._performCommand("repr", item)
                return
            if key == QtCore.Qt.Key.Key_Backspace:
                self._proxy.goUp()
                return
        super().keyPressEvent(event)

    def contextMenuEvent(self, event):
        """Show the context menu."""

        QtWidgets.QTreeView.contextMenuEvent(self, event)

        # Get if an item is selected
        item = self.currentItem()
        if not item:
            return

        # Create menu
        self._menu.clear()
        commands = [
            ("Show namespace", pyzo.translate("pyzoWorkspace", "Show namespace")),
            ("Show help", pyzo.translate("pyzoWorkspace", "Show help")),
            (
                "Add to Expr Viewer",
                pyzo.translate("pyzoWorkspace", "Add to Expression viewer tool"),
            ),
            ("str", "print(variable)"),
            ("repr", "print(repr(variable))"),
            ("Delete", pyzo.translate("pyzoWorkspace", "Delete")),
        ]
        for request, display in commands:
            action = self._menu.addAction(display)
            action._what = request
            action._item = item

        # Show
        self._menu.popup(QtGui.QCursor.pos() + QtCore.QPoint(3, 3))

    def _contextMenuTriggered(self, action):
        """Process a request from the context menu."""
        self._performCommand(action._what, action._item)

    def _performCommand(self, cmd, item):
        objectName = joinName(splitName(self._proxy._name) + [item.text(0)])
        if cmd == "Show namespace":
            # Go deeper
            self.onItemExpand(item)
        elif cmd == "Show help":
            # Show help in help tool (if loaded)
            hw = pyzo.toolManager.getTool("pyzointeractivehelp")
            if hw:
                hw.setObjectName(objectName, addToHist=True)
        elif cmd == "Add to Expr Viewer":
            self._addExpressionToViewer(objectName)
        elif cmd in ("str", "repr"):
            shell = pyzo.shells.getCurrentShell()
            if shell is not None:
                if cmd == "str":
                    shell.executeCommand("print({})\n".format(objectName))
                else:
                    shell.executeCommand("print(repr({}))\n".format(objectName))
        elif cmd == "Delete":
            # Delete the variable
            shell = pyzo.shells.getCurrentShell()
            if shell is not None:
                shell.processLine("del " + objectName)

    def _addExpressionToViewer(self, expr, doRefresh=True):
        """add an expression to Expression viewer tool (if loaded)"""
        ev = pyzo.toolManager.getTool("pyzoexpressionviewer")
        if ev:
            ev.addExpression(expr)
            ev.manualRefresh()

    def onItemExpand(self, item):
        """Inspect the attributes of that item."""
        self._proxy.addNamePart(item.text(0))

    def setCompiledRegExp(self, compiledRegExp_or_errorstring):
        """Set the name filter for variables via a compiled regular expression."""
        self._compiledRegExp = compiledRegExp_or_errorstring

    def fillWorkspace(self):
        """Update the workspace tree."""

        try:
            selectedName = self.selectedItems()[0].text(0)
        except Exception:
            selectedName = None
        newSelectedItem = None

        if isinstance(self._compiledRegExp, str):
            errorMessage = self._compiledRegExp
            self.parent().displayEmptyWorkspace(True, errorMessage)
            return

        # Set name
        line = self.parent()._line
        line.setText(self._proxy._name)

        self.setSortingEnabled(False)

        # Add elements
        # The widget is not cleared to keep its current scroll bar position.
        i = -1
        for name, typeName, kind, repres in self._proxy._variables:
            # <kludge 2>
            # the typeTranslation dictionary contains "synonyms" for types that will be hidden
            kind = self._config.typeTranslation.get(kind, kind)
            # </kludge 2>
            if kind in self._config.hideTypes:
                continue
            if name.startswith("_") and "private" in self._config.hideTypes:
                continue
            if "startup" in self._config.hideTypes and name in self._startUpVariables:
                continue
            if not self._compiledRegExp.fullmatch(name):
                continue

            i += 1
            item = self.topLevelItem(i)
            if item is None:
                # Create item
                item = WorkspaceItem((name, typeName, repres), 0)
                self.addTopLevelItem(item)
            else:
                item.setText(0, name)
                item.setText(1, typeName)
                item.setText(2, repres)
                item.setSelected(False)

            if name == selectedName:
                newSelectedItem = item

            # Set tooltip
            tt = "{}: {}".format(name, repres)
            item.setToolTip(0, tt)
            item.setToolTip(1, tt)
            item.setToolTip(2, tt)

        for _ in range(i + 1, self.topLevelItemCount()):
            self.takeTopLevelItem(i + 1)  # delete remaining entries

        self.setSortingEnabled(True)
        if newSelectedItem is not None:
            newSelectedItem.setSelected(True)
            self.setCurrentItem(newSelectedItem)

        self.parent().displayEmptyWorkspace(
            self.topLevelItemCount() == 0 and self._proxy._name == ""
        )

    def _addAllToViewer(self):
        """add all expressions to the Expression viewer tool (if loaded)"""
        n = self.topLevelItemCount()
        for i in range(n):
            item = self.topLevelItem(i)
            if item is not None:
                objectName = joinName(splitName(self._proxy._name) + [item.text(0)])
                self._addExpressionToViewer(objectName, doRefresh=(i == n - 1))


class PyzoWorkspace(QtWidgets.QWidget):
    """PyzoWorkspace

    The main widget for this tool.

    """

    def __init__(self, parent):
        super().__init__(parent)

        # Make sure there is a configuration entry for this tool
        # The pyzo tool manager makes sure that there is an entry in
        # config.tools before the tool is instantiated.
        toolId = self.__class__.__name__.lower()
        self._config = pyzo.config.tools[toolId]
        self._config.setdefault("hideTypes", [])
        self._config.setdefault(
            "typeTranslation",
            {
                "method": "function",
                "builtin_function_or_method": "function",
            },
        )

        # Create tool buttons
        self._refresh = QtWidgets.QToolButton(self)
        self._refresh.setIcon(pyzo.icons.arrow_refresh)
        self._refresh.setIconSize(QtCore.QSize(16, 16))
        self._refresh.setToolTip(
            "manually refresh variables (e.g. for code running in the event loop)\n"
            "this will also refresh the Expression viewer tool, if open"
        )

        self._up = QtWidgets.QToolButton(self)
        self._up.setIcon(pyzo.icons.arrow_left)
        self._up.setIconSize(QtCore.QSize(16, 16))
        self._up.setToolTip("switch to parent namespace")

        self._btnAddAllToViewer = QtWidgets.QToolButton(self)
        self._btnAddAllToViewer.setIcon(pyzo.icons.magnifier_zoom_in)
        self._btnAddAllToViewer.setIconSize(QtCore.QSize(16, 16))
        self._btnAddAllToViewer.setToolTip(
            "add all visible entries to the Expression viewer tool"
        )

        # Create "path" line edit
        self._line = QtWidgets.QLineEdit(self)
        self._line.setReadOnly(True)
        self._line.setStyleSheet(
            "QLineEdit {{ background:{}; }}".format("#888" if pyzo.darkQt else "#ddd")
        )
        self._line.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        # Create options menu
        self._options = QtWidgets.QToolButton(self)
        self._options.setIcon(pyzo.icons.filter)
        self._options.setIconSize(QtCore.QSize(16, 16))
        self._options.setPopupMode(self._options.ToolButtonPopupMode.InstantPopup)
        self._options.setToolButtonStyle(
            QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon
        )
        #
        self._options._menu = QtWidgets.QMenu()
        self._options.setMenu(self._options._menu)
        self._onOptionsPress()  # create menu now

        # Create tree
        self._tree = WorkspaceTree(self)

        # Create message for when tree is empty
        self._no_results_text = pyzo.translate(
            "pyzoWorkspace",
            "Lists the variables in the current shell's namespace."
            "\n\n"
            "Currently, there are none. Some of them may be hidden because of the filters you configured.",
        )
        self._noResultsLabel = QtWidgets.QLabel(self._no_results_text, self)
        self._noResultsLabel.setVisible(False)
        self._noResultsLabel.setWordWrap(True)

        # Create search line edit
        self._searchText = QtWidgets.QLineEdit(self)

        # Create search options button
        self._searchOptions = QtWidgets.QToolButton(self)
        self._searchOptions.setIcon(pyzo.icons.magnifier)
        self._searchOptions.setIconSize(QtCore.QSize(16, 16))
        self._searchOptions.setPopupMode(
            self._searchOptions.ToolButtonPopupMode.InstantPopup
        )
        self._searchOptions.setToolButtonStyle(
            QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon
        )
        #
        self._searchOptions._menu = QtWidgets.QMenu()
        self._searchOptions.setMenu(self._searchOptions._menu)
        self._onSearchOptionsPress()  # create menu now and init default config values
        self._updateSearchPlaceHolderText()

        # We put the tooltip text to the search options button instead of the
        # search text widget to make it less annoying.
        self._searchOptions.setToolTip(
            "Enter one or multiple space-separated search expressions.\n"
            "A variable name will be displayed if it matches at least one of the expressions.\n"
            "\n"
            "Depending on the settings, search expressions are either\n"
            "\n"
            "texts with wildcards:\n"
            "    a single question mark (?) matches exactly one character\n"
            "    an asterisk (*) matches zero or multiple characters\n"
            '    example: "a* b*" ... matches all variables beginning with "a" or "b"\n'
            "\n"
            "or regular expressions:\n"
            '    example: "myvar[0-5].* obj\\d+ x.*"'
        )

        # Set layout
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._refresh, 0)
        layout.addWidget(self._up, 0)
        layout.addWidget(self._line, 1)
        layout.addWidget(self._btnAddAllToViewer, 0)
        layout.addWidget(self._options, 0)
        #
        searchLayout = QtWidgets.QHBoxLayout()
        searchLayout.addWidget(self._searchText, 1)
        searchLayout.addWidget(self._searchOptions, 0)
        #
        mainLayout = QtWidgets.QVBoxLayout(self)
        mainLayout.addLayout(layout, 0)
        mainLayout.addWidget(self._noResultsLabel, 1)
        mainLayout.addWidget(self._tree, 2)
        mainLayout.setSpacing(2)
        mainLayout.addLayout(searchLayout, 0)

        # set margins
        margin = pyzo.config.view.widgetMargin
        mainLayout.setContentsMargins(margin, margin, margin, margin)
        self.setLayout(mainLayout)

        # Bind events
        self._up.pressed.connect(self._tree._proxy.goUp)
        self._refresh.pressed.connect(self._onButtonRefresh)
        self._options.pressed.connect(self._onOptionsPress)
        self._options._menu.triggered.connect(self._onOptionMenuTiggered)
        self._searchOptions.pressed.connect(self._onSearchOptionsPress)
        self._searchOptions._menu.triggered.connect(self._onSearchOptionMenuTiggered)
        self._searchText.textChanged.connect(self._onSearchTextUpdated)
        self._btnAddAllToViewer.pressed.connect(self._tree._addAllToViewer)

    def displayEmptyWorkspace(self, empty, customMessage=None):
        if customMessage is None:
            customMessage = self._no_results_text
        self._tree.setVisible(not empty)
        self._noResultsLabel.setText(customMessage)
        self._noResultsLabel.setVisible(empty)

    def _onButtonRefresh(self):
        self.manualRefresh()
        ev = pyzo.toolManager.getTool("pyzoexpressionviewer")
        if ev:
            ev.manualRefresh()

    def manualRefresh(self):
        """manually refresh the variables"""
        self._tree._proxy.updateVariables()

    def _onOptionsPress(self):
        """Create the menu for the button, Do each time to make sure
        the checks are right."""

        # Get menu
        menu = self._options._menu
        menu.clear()

        hideables = [
            ("type", pyzo.translate("pyzoWorkspace", "Hide types")),
            ("function", pyzo.translate("pyzoWorkspace", "Hide functions")),
            ("module", pyzo.translate("pyzoWorkspace", "Hide modules")),
            ("private", pyzo.translate("pyzoWorkspace", "Hide private identifiers")),
            (
                "startup",
                pyzo.translate("pyzoWorkspace", "Hide the shell's startup variables"),
            ),
        ]

        for type, display in hideables:
            checked = type in self._config.hideTypes
            action = menu.addAction(display)
            action._what = type
            action.setCheckable(True)
            action.setChecked(checked)

    def _onOptionMenuTiggered(self, action):
        """The user decides what to hide in the workspace."""

        # What to show
        type = action._what.lower()

        self._config.hideTypes = list(set(self._config.hideTypes) ^ {type})

        # Update
        self._tree.fillWorkspace()

    def _onSearchOptionsPress(self):
        """Create the menu for the button, Do each time to make sure
        the checks are right."""

        menu = self._searchOptions._menu
        menu.clear()

        searchOptions = [
            ("searchMatchCase", True, pyzo.translate("pyzoWorkspace", "Match case")),
            ("searchRegExp", False, pyzo.translate("pyzoWorkspace", "RegExp")),
            (
                "searchStartsWith",
                True,
                pyzo.translate("pyzoWorkspace", "Starts with ..."),
            ),
        ]

        for name, default, label in searchOptions:
            action = menu.addAction(label)
            action._what = name
            action.setCheckable(True)
            action.setChecked(self._config.setdefault(name, default))

    def _onSearchOptionMenuTiggered(self, action):
        name = action._what
        self._config[name] ^= True

        # Update
        self._onSearchTextUpdated()
        self._tree.fillWorkspace()
        self._updateSearchPlaceHolderText()

    def _onSearchTextUpdated(self):
        needles = self._searchText.text().split()
        flags = re.IGNORECASE if not self._config.searchMatchCase else 0
        if len(needles) == 0:
            regExpList = [r".*"]
        elif self._config.searchRegExp:
            if self._config.searchStartsWith:
                needles = [s + ".*" for s in needles]
            regExpList = needles
        else:
            if self._config.searchStartsWith:
                needles = [s + "*" for s in needles]
            regExpList = [wildcardsToRegExp(s) for s in needles]

        pattern = "(?:" + ")|(?:".join(regExpList) + ")"
        try:
            compiledRegExp = re.compile(pattern, flags)
        except Exception as e:
            compiledRegExp = "invalid regular expression:\n" + pattern + "\n" + str(e)

        self._tree.setCompiledRegExp(compiledRegExp)
        self._tree.fillWorkspace()

    def _updateSearchPlaceHolderText(self):
        # Update
        self._searchText.setPlaceholderText(
            "[{} mode, case {}]".format(
                "RegExp" if self._config.searchRegExp else "wildcards",
                "sensitive" if self._config.searchMatchCase else "insensitive",
            )
        )



import re
import os
import sys
import subprocess

from pyzo.qt import QtCore


def parse_version_crudely(version_string):
    """extracts the leading number parts of a version string to a tuple
    e.g.: "123.45ew6.7x.dev8" --> (123, 45, 7)
    """
    return tuple(int(s) for s in re.findall(r"\.(\d+)", "." + version_string))


def open_directory_outside_pyzo(dirpath, filename=None):
    """open the directory in the operating system's file browser.

    If possible, also select a specific filename (only supported in MS Windows).
    """
    if not os.path.isdir(dirpath):
        return

    if filename is not None:
        filepath = os.path.join(dirpath, filename)
        if not os.path.isfile(filepath):
            filepath = None
    else:
        filepath = None

    if sys.platform.startswith("darwin"):
        subprocess.call(("open", dirpath))
    elif sys.platform.startswith("win"):
        if filepath is not None:
            subprocess.call('explorer.exe /select,"{}"'.format(filepath))
        else:
            subprocess.call('explorer.exe "{}"'.format(dirpath))
    elif sys.platform.startswith("linux"):
        subprocess.call(("xdg-open", dirpath))


class CalmedFunc:
    """Class for delayed execution of a function so that newer calls will replace the
    not yet executed older calls.

    If there is no newer call within a certain amount of time ("delay_ms"), the function
    will be executed with the arguments from the newest call.

    The first call after a longer break (after "delay_ms" or after initialization)
    will be executed immediately. The next execution will not happen before "delay_ms"
    since the previous execution, unless method "clear_pending" is called before.
    """

    def __init__(self, func, delay_ms):
        self._func = func
        self._t = QtCore.QTimer(singleShot=True)
        self._t.timeout.connect(self._timer_callback)
        self._quick_start = False
        self._wait_after_quick_start = False
        self.set_delay(delay_ms)

    def set_delay(self, delay_ms):
        self._delay_ms = delay_ms

    def clear_pending(self):
        self._t.stop()

    def __call__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._wait_after_quick_start = False
        if self._t.isActive():
            self._t.start(self._delay_ms)
        else:
            self._quick_start = True
            self._t.start(1)

    def _timer_callback(self):
        if self._wait_after_quick_start:
            self._wait_after_quick_start = False
            return

        self._func(*self._args, **self._kwargs)

        if self._quick_start:
            self._wait_after_quick_start = True
            self._t.start(self._delay_ms)



# This is an example file for the "Snippets" tool.
# If there is no "snippets" folder in the Pyzo userdata directory, Pyzo will automatically
# create that folder and copy this file there, just as an example to get started.
# To obtain the Pyzo userdata directory, see the "About Pyzo" dialog ("Help -> About Pyzo"
# in the menu).

# You can put multiple *.py files into the snippets directory.
# It is even possible to create namespace groups by placing files into nested subfolders.
# Snippet files must have UTF-8 encoding (or ASCII, which is a subset of UTF-8).
# Snippet and subfolder names must be valid ASCII identifiers: "[a-zA-Z_][a-zA-Z_0-9]*"
# Snippet names of all snippets in the same directory (without subfolders) must be unique.

# Every snippet starts with a cell comment line, such as "## SNIP: name_of_the_snippet".
# All comment lines directly below the cell comment line are an optional description
# of the snippet. This description will be shown in the tooltip text in the Snippets tool
# as well as in the "Interactive help" tool, but it will not be inserted with the code.

# On insertion, the snippet code will be inserted at the current location of the text
# cursor in the active editor tab. If there is only whitespace before the cursor position,
# the snippet code lines will be prefixed by the same whitespace characters to have the
# same indentation.

# Empty lines before and after the snippet will be removed when reading the file.
# Snippets with only one code line will be inserted without a linebreak at the end.


## SNIP: readme
# This is an example snippet.
# In the Pyzo "Snippets" tool, right click on the list entry and select "Edit" to open
# this file.
# All comment lines below the snippet header are part of the description.
# The description stops on the first non-comment line, such as the following empty line.

# This line is already part of the snippet code.

import antigravity


## SNIP: changepath
import inspect
import os

__this_dir__ = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))
os.chdir(__this_dir__)


## SNIP: changepath_and_modulespath
# change path to path of current script
# this also works when executing code as cell (contrary to __file__)

import inspect
import os
import sys

__this_dir__ = os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe())))
os.chdir(__this_dir__)

__my_modules_path__ = os.path.join(__this_dir__, r'my_modules')
if __my_modules_path__ not in sys.path:
    sys.path.append(__my_modules_path__)


## SNIP: plot_with_imports
import numpy as np
import matplotlib.pyplot as plt
plt.ion()

fig, ax = plt.subplots(num=10, clear=True)
ax.plot([1, 2, 4], '-x')
ax.grid(True)


## SNIP: lowpass
# digital lowpass filter, bessel, 2nd order
# with cutoff frequency fc and sample frequency fs

import scipy.signal

filter_b_a = scipy.signal.bessel(2, fc/(fs/2), 'lowpass')
sig_f = scipy.signal.lfilter(*filter_b_a, sig)


## SNIP: time_elapsed
from time import perf_counter as _get_time
# from time import time as _get_time
t_start = _get_time()
...
print('elapsed:', t_delta := _get_time() - t_start, 's')


## SNIP: Qt_import_PySide6
from PySide6 import QtCore, QtGui, QtWidgets


## SNIP: walk_files
import os

basepath = r''
for dirpath, dirnames, filenames in os.walk(basepath):
    dirnames[:] = [dn for dn in dirnames if not dn.startswith('.')]
    for fn in filenames:
        if fn.endswith('.txt'):
            fp = os.path.join(dirpath, fn)
            print(fp)
            ...




"""Module highlighter

Defines the highlighter class for the base code editor class. It will do
the styling when syntax highlighting is enabled. If it is not, will only
check out indentation.

"""

from .qt import QtGui, QtCore

Qt = QtCore.Qt

from . import parsers


class BlockData(QtGui.QTextBlockUserData):
    """Class to represent the data for a block."""

    def __init__(self):
        super().__init__()
        self.indentation = None
        self.fullUnderlineFormat = None
        self.tokens = []


# The highlighter should be part of the base class, because
# some extensions rely on them (e.g. the indent guides).
class Highlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, codeEditor, *args):
        super().__init__(*args)

        # Store reference to editor
        self._codeEditor = codeEditor

    def getCurrentBlockUserData(self):
        """Gets the BlockData object. Creates one if necessary."""
        bd = self.currentBlockUserData()
        if not isinstance(bd, BlockData):
            bd = BlockData()
            self.setCurrentBlockUserData(bd)
        return bd

    def highlightBlock(self, line):
        """This method is automatically called when a line must be re-highlighted.

        If the code editor has an active parser. This method will use
        it to perform syntax highlighting. If not, it will only
        check out the indentation.
        """

        # Get previous state
        previousState = self.previousBlockState()

        # Get parser
        parser = None
        if hasattr(self._codeEditor, "parser"):
            parser = self._codeEditor.parser()

        # Get function to get format
        nameToFormat = self._codeEditor.getStyleElementFormat

        fullLineFormat = None
        tokens = []
        if parser:
            self.setCurrentBlockState(0)
            tokens = parser.parseLine(line, previousState)
            for token in tokens:
                # Handle block state
                if isinstance(token, parsers.BlockState):
                    self.setCurrentBlockState(token.state)
                else:
                    # Get format
                    try:
                        styleFormat = nameToFormat(token.name)
                        charFormat = styleFormat.textCharFormat
                    except KeyError:
                        # print(repr(nameToFormat(token.name)))
                        continue
                    # Set format
                    self.setFormat(token.start, token.end - token.start, charFormat)
                    # Is this a cell?
                    if (fullLineFormat is None) and styleFormat._parts.get(
                        "underline", ""
                    ) == "full":
                        fullLineFormat = styleFormat

        # Get user data
        bd = self.getCurrentBlockUserData()

        # Store token list for future use (e.g. brace matching)
        bd.tokens = tokens

        # Handle underlines
        bd.fullUnderlineFormat = fullLineFormat

        # Get the indentation setting of the editors
        indentUsingSpaces = self._codeEditor.indentUsingSpaces()

        leadingWhitespace = line[: len(line) - len(line.lstrip())]
        US = QtGui.QTextCharFormat.UnderlineStyle
        if "\t" in leadingWhitespace and " " in leadingWhitespace:
            # Mixed whitespace
            bd.indentation = 0
            format = QtGui.QTextCharFormat()
            format.setUnderlineStyle(US.SpellCheckUnderline)
            format.setUnderlineColor(QtCore.Qt.GlobalColor.red)
            format.setToolTip("Mixed tabs and spaces")
            self.setFormat(0, len(leadingWhitespace), format)
        elif ("\t" in leadingWhitespace and indentUsingSpaces) or (
            " " in leadingWhitespace and not indentUsingSpaces
        ):
            # Whitespace differs from document setting
            bd.indentation = 0
            format = QtGui.QTextCharFormat()
            format.setUnderlineStyle(US.SpellCheckUnderline)
            format.setUnderlineColor(QtCore.Qt.GlobalColor.blue)
            format.setToolTip("Whitespace differs from document setting")
            self.setFormat(0, len(leadingWhitespace), format)
        else:
            # Store info for indentation guides
            # amount of tabs or spaces
            bd.indentation = len(leadingWhitespace)
